""" Test methods for Connection """
import io
from copy import deepcopy
from unittest.mock import MagicMock, patch

import pytest

from qiboconnection import __version__
from qiboconnection.connection import Connection, refresh_token_if_unauthorised
from qiboconnection.errors import HTTPError, RemoteExecutionException
from qiboconnection.models.user import User

from .data import web_responses


def test_constructor(mocked_connection: Connection):
    """Test Connection class constructor"""
    assert isinstance(mocked_connection, Connection)


def test_user(mocked_connection: Connection):
    """Test Connection user property"""
    assert isinstance(mocked_connection.user, User)


def test_user_when_not_defined(mocked_connection_no_user: Connection):
    """Test Connection user property"""

    with pytest.raises(ValueError):
        _ = mocked_connection_no_user.user


def test_username(mocked_connection: Connection):
    """Test Connection user property"""
    assert isinstance(mocked_connection.username, str)


def test_username_when_user_not_defined(mocked_connection_no_user: Connection):
    """Test Connection user property"""

    with pytest.raises(ValueError):
        _ = mocked_connection_no_user.username


def test_user_id(mocked_connection: Connection):
    """Test Connection user property"""
    assert isinstance(mocked_connection._user_id, int)


def test_connection_user_id_when_user_not_defined(mocked_connection_no_user: Connection):
    """Test Connection user property"""
    with pytest.raises(ValueError):
        _ = mocked_connection_no_user._user_id


def test_connection_user_id_setter(mocked_connection: Connection):
    """Test Connection user property"""
    mocked_connection._user_id = 555
    assert mocked_connection._user_id == 555


def test_connection_user_id_setter_when_user_not_defined(mocked_connection_no_user: Connection):
    """Test Connection user property"""
    with pytest.raises(ValueError):
        mocked_connection_no_user._user_id = 555


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_user_slack_id_local(mocked_web_call: MagicMock, mocked_connection: Connection):
    """Test Connection user_slack_id property"""
    mocked_web_call.return_value = web_responses.users.retrieve_response
    connection = deepcopy(mocked_connection)
    connection._user_slack_id = "Random ID"

    slack_id = connection.user_slack_id

    mocked_web_call.assert_not_called()
    assert slack_id == "Random ID"


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_user_slack_id(mocked_web_call: MagicMock, mocked_connection: Connection):
    """Test Connection user_slack_id property"""
    mocked_web_call.return_value = web_responses.users.retrieve_response
    connection = deepcopy(mocked_connection)
    connection._user_slack_id = None

    slack_id = connection.user_slack_id

    mocked_web_call.assert_called_with(self=connection, path=f"/users/{connection.user.user_id}")
    assert slack_id == web_responses.users.retrieve_response[0]["slack_id"]


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_user_slack_id_without_slack_id(mocked_web_call: MagicMock, mocked_connection: Connection):
    """Test Connection user_slack_id property"""
    mocked_web_call.return_value = web_responses.users.retrieve_response_without_slack_id
    connection = deepcopy(mocked_connection)
    connection._user_slack_id = None

    slack_id = connection.user_slack_id

    mocked_web_call.assert_called_with(self=connection, path=f"/users/{connection.user.user_id}")
    assert slack_id == ""


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_user_slack_id_ise(mocked_web_call: MagicMock, mocked_connection: Connection):
    """Test Connection user_slack_id property fails with failing response"""
    mocked_web_call.return_value = web_responses.users.ise_response
    connection = deepcopy(mocked_connection)

    with pytest.raises(ValueError):
        _ = connection.user_slack_id

    mocked_web_call.assert_called_with(self=connection, path=f"/users/{connection.user.user_id}")


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_user_slack_id_without_user(mocked_web_call: MagicMock, mocked_connection_no_user: Connection):
    """Test Connection user_slack_id property fails without local user"""
    mocked_web_call.return_value = web_responses.users.retrieve_response

    with pytest.raises(ValueError):
        _ = mocked_connection_no_user.user_slack_id

    mocked_web_call.assert_not_called()


@patch("qiboconnection.connection.requests.put", autospec=True)
def test_send_put_auth_remote_api_call(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """tests send_put_auth_remote_api_call"""
    mocked_rest_call.return_value = web_responses.raw.response_200

    response, code = mocked_connection.send_put_auth_remote_api_call(path="/PATH", data={"demo": "demo"}, timeout=10)

    mocked_rest_call.assert_called_with(
        f"{mocked_connection._remote_server_api_url}/PATH",
        json={"demo": "demo"},
        headers={
            "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
            ".eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ"
            ".SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c",
            "X-Client-Version": __version__,
        },
        timeout=10,
    )
    assert response == web_responses.raw.response_200.json()
    assert code == web_responses.raw.response_200.status_code


@patch("qiboconnection.connection.requests.post", autospec=True)
def test_send_post_auth_remote_api_call(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """tests send_post_auth_remote_api_call"""
    mocked_rest_call.return_value = web_responses.raw.response_201

    response, code = mocked_connection.send_post_auth_remote_api_call(path="/PATH", data={"demo": "demo"}, timeout=10)

    mocked_rest_call.assert_called_with(
        f"{mocked_connection._remote_server_api_url}/PATH",
        json={"demo": "demo"},
        headers={
            "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
            ".eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ"
            ".SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c",
            "X-Client-Version": __version__,
        },
        timeout=10,
    )
    assert response == web_responses.raw.response_201.json()
    assert code == web_responses.raw.response_201.status_code


@patch("qiboconnection.connection.requests.get", autospec=True)
def test_send_get_auth_remote_api_call(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """tests send_get_auth_remote_api_call"""
    mocked_rest_call.return_value = web_responses.raw.response_200

    response, code = mocked_connection.send_get_auth_remote_api_call(path="/PATH", params={"demo": "demo"}, timeout=10)

    mocked_rest_call.assert_called_with(
        f"{mocked_connection._remote_server_api_url}/PATH",
        params={"demo": "demo"},
        headers={
            "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
            ".eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ"
            ".SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c",
            "X-Client-Version": __version__,
        },
        timeout=10,
    )
    assert response == web_responses.raw.response_200.json()
    assert code == web_responses.raw.response_200.status_code


@patch("qiboconnection.connection.requests.get", autospec=True)
def test_send_get_auth_remote_api_call_exception(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """tests send_get_auth_remote_api_call"""

    mocked_rest_call.return_value = web_responses.job_response.retrieve_job_response_400_raw()

    with pytest.raises(RemoteExecutionException, match="The job does not exist!"):
        # Call the function that should raise the exception
        mocked_connection.send_get_auth_remote_api_call(path="/PATH", params={"demo": "demo"})


@patch("qiboconnection.connection.requests.delete", autospec=True)
def test_send_delete_auth_remote_api_call(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """tests send_delete_auth_remote_api_call"""
    mocked_rest_call.return_value = web_responses.raw.response_204

    response, code = mocked_connection.send_delete_auth_remote_api_call(path="/PATH", timeout=10)

    mocked_rest_call.assert_called_with(
        f"{mocked_connection._remote_server_api_url}/PATH",
        headers={
            "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
            ".eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ"
            ".SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c",
            "X-Client-Version": __version__,
        },
        timeout=10,
    )
    assert response == web_responses.raw.response_204.json()
    assert code == web_responses.raw.response_204.status_code


@patch("qiboconnection.connection.requests.delete", autospec=True)
def test_send_delete_auth_remote_api_call_not_204_not_job_details(
    mocked_rest_call: MagicMock, mocked_connection: Connection
):
    """tests send_delete_auth_remote_api_call raise"""
    mocked_rest_call.return_value = web_responses.raw.response_500
    with pytest.raises(HTTPError):
        _, _ = mocked_connection.send_delete_auth_remote_api_call(path="/PATH", timeout=10)

    mocked_rest_call.assert_called_with(
        f"{mocked_connection._remote_server_api_url}/PATH",
        headers={
            "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
            ".eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ"
            ".SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c",
            "X-Client-Version": __version__,
        },
        timeout=10,
    )


@patch("qiboconnection.connection.requests.delete", autospec=True)
def test_send_delete_auth_remote_api_call_not_204_with_job_details(
    mocked_rest_call: MagicMock, mocked_connection: Connection
):
    """tests send_delete_auth_remote_api_call raise"""
    mocked_rest_call.return_value = web_responses.job_response.retrieve_job_response_400_raw()
    with pytest.raises(RemoteExecutionException, match="The job does not exist!"):
        _, _ = mocked_connection.send_delete_auth_remote_api_call(path="/PATH", timeout=10)

    mocked_rest_call.assert_called_with(
        f"{mocked_connection._remote_server_api_url}/PATH",
        headers={
            "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
            ".eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ"
            ".SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c",
            "X-Client-Version": __version__,
        },
        timeout=10,
    )


@patch("qiboconnection.connection.requests.post", autospec=True)
def test_send_post_file_auth_remote_api_call(mocked_rest_call: MagicMock, mocked_connection: Connection):
    mocked_rest_call.return_value = web_responses.raw.response_200

    _PATH = "/PATH"
    _FILE = io.StringIO()
    _FILENAME = "filename"
    _TIMEOUT = 10

    mocked_connection.send_post_file_auth_remote_api_call(path=_PATH, file=_FILE, filename=_FILENAME, timeout=_TIMEOUT)

    mocked_rest_call.assert_called_with(
        f"{mocked_connection._remote_server_api_url}{_PATH}",
        files={"file": (_FILENAME, _FILE)},
        headers={
            "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
            ".eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ"
            ".SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c",
            "X-Client-Version": __version__,
        },
        timeout=_TIMEOUT,
    )


@patch("qiboconnection.connection.requests.get", autospec=True)
def test_send_get_auth_remote_api_call_all_pages(mocked_rest_call: MagicMock, mocked_connection: Connection):
    mocked_rest_call.side_effect = [
        web_responses.raw.response_200_paginated,
        web_responses.raw.response_200_paginated_last,
    ]

    _PATH = "/path"
    _PARAMS = {"param": "value"}
    _TIMEOUT = 10

    responses = mocked_connection.send_get_auth_remote_api_call_all_pages(path=_PATH, params=_PARAMS, timeout=_TIMEOUT)

    mocked_rest_call.assert_called()
    assert responses == [
        (
            {
                "items": [{"key": "value"}],
                "total": 2,
                "per_page": 5,
                "self": "https://qilimanjarodev.ddns.net:8080/api/v1/path?page=1&per_page=5",
                "links": {
                    "first": "https://qilimanjarodev.ddns.net:8080/api/v1/path?page=1&per_page=5",
                    "prev": "https://qilimanjarodev.ddns.net:8080/api/v1/path?page=None&per_page=5",
                    "next": "https://qilimanjarodev.ddns.net:8080/api/v1/path?page=None&per_page=5",
                    "last": "https://qilimanjarodev.ddns.net:8080/api/v1/path?page=1&per_page=5",
                },
            },
            200,
        )
    ]


@patch("qiboconnection.connection.requests.get", autospec=True)
def test_send_get_remote_call(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """tests send_get_remote_call"""
    mocked_rest_call.return_value = web_responses.raw.response_200

    response, code = mocked_connection.send_get_remote_call(path="/PATH", timeout=10)

    mocked_rest_call.assert_called_with(
        f"{mocked_connection._remote_server_base_url}/PATH", timeout=10, headers={"X-Client-Version": __version__}
    )
    assert response == web_responses.raw.response_200.json()
    assert code == web_responses.raw.response_200.status_code


@patch("qiboconnection.connection.requests.post", autospec=True)
def test_request_authorisation_token_with_ok_response(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """tests send_get_remote_call"""
    mocked_rest_call.return_value = web_responses.auth.raw_retrieve_response

    access_token, refresh_token = mocked_connection._request_authorisation_token()

    assert (
        access_token == web_responses.auth.retrieve_response[0]["accessToken"]
    ), "accessToken does not coincide with expected"

    assert (
        refresh_token == web_responses.auth.retrieve_response[0]["refreshToken"]
    ), "refreshToken does not coincide with expected"


@patch("qiboconnection.connection.requests.post", autospec=True)
def test_request_authorisation_token_with_no_url_raises_error(
    mocked_rest_call: MagicMock, mocked_connection: Connection
):
    """tests send_get_remote_call"""

    mocked_connection_copy = deepcopy(mocked_connection)
    mocked_connection_copy._authorisation_server_api_call = None

    with pytest.raises(ValueError, match="Authorisation server api call is required"):
        _, _ = mocked_connection_copy._request_authorisation_token()
        mocked_rest_call.assert_not_called()


@patch("qiboconnection.connection.requests.post", autospec=True)
def test_request_authorisation_token_with_ise(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """tests send_get_remote_call"""
    mocked_rest_call.return_value = web_responses.raw.response_500

    with pytest.raises(ValueError, match="Authorisation request failed."):
        _, _ = mocked_connection._request_authorisation_token()


@patch("qiboconnection.connection.requests.post", autospec=True)
def test_update_authorisation_using_refresh_token_no_server_refresh_api_call(
    mocked_rest_call: MagicMock, mocked_connection: Connection
):
    """test Connection's update_authorisation_using_refresh_token, with no refresh api call defined"""

    mocked_connection_copy = deepcopy(mocked_connection)

    mocked_rest_call.return_value = web_responses.auth.raw_retrieve_response

    mocked_connection_copy._authorisation_server_refresh_api_call = None

    with pytest.raises(ValueError, match="Authorisation server api call is required"):
        mocked_connection_copy.update_authorisation_using_refresh_token()


@patch("qiboconnection.connection.requests.post", autospec=True)
def test_update_authorisation_using_refresh_token_unsuccessful(
    mocked_rest_call: MagicMock, mocked_connection: Connection
):
    """test Connection's update_authorisation_using_refresh_token"""

    mocked_rest_call.return_value = web_responses.raw.response_400

    mocked_connection._authorisation_refresh_token = (
        "eyJhbGciOiJFZERTQSIsImtpZCI6InlYaG5lSUtxUEV5UklSLXVyMHdGZUZzLTZ2VS01amJEY18wUFN0X2Etc1UiLCJ0eXAiOiJKV1QifQ"
        + ".eyJhdWQiOiJodHRwczovL3FpbGltYW5qYXJvZGV2LmRkbnMubmV0OjgwODAvYXBpL3YxIiwiZXhwIjoxNjg0NDk"
        + "yMDcxLCJpYXQiOjE2ODQ0MDU2NzEsImlzcyI6Imh0dHBzOi8vcWlsaW1hbmphcm9kZXYuZGRucy5uZXQ6ODA4MC8i"
        + "LCJ0eXBlIjoicmVmcmVzaCIsInVzZXJfaWQiOjMsInVzZXJfcm9sZSI6ImFkbWluIn0"
        + ".4oSyRW9Ia7C-50x2yZxQAEXDZp-TLkFkPOtHBR4cCi9LnkREtYrJpDXufep_EYoRwDSJL_2z20moYMuMHy0QCg"
    )

    with pytest.raises(ValueError, match=f"Authorisation request failed: {web_responses.raw.response_400.reason}"):
        mocked_connection.update_authorisation_using_refresh_token()


@patch("qiboconnection.connection.requests.post", autospec=True)
def test_update_authorisation_using_refresh_token(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """test Connection's update_authorisation_using_refresh_token"""

    mocked_rest_call.return_value = web_responses.auth.raw_retrieve_response

    mocked_connection._authorisation_refresh_token = (
        "eyJhbGciOiJFZERTQSIsImtpZCI6InlYaG5lSUtxUEV5UklSLXVyMHdGZUZzLTZ2VS01amJEY18wUFN0X2Etc1UiLCJ0eXAiOiJKV1QifQ"
        + ".eyJhdWQiOiJodHRwczovL3FpbGltYW5qYXJvZGV2LmRkbnMubmV0OjgwODAvYXBpL3YxIiwiZXhwIjoxNjg0NDk"
        + "yMDcxLCJpYXQiOjE2ODQ0MDU2NzEsImlzcyI6Imh0dHBzOi8vcWlsaW1hbmphcm9kZXYuZGRucy5uZXQ6ODA4MC8i"
        + "LCJ0eXBlIjoicmVmcmVzaCIsInVzZXJfaWQiOjMsInVzZXJfcm9sZSI6ImFkbWluIn0"
        + ".4oSyRW9Ia7C-50x2yZxQAEXDZp-TLkFkPOtHBR4cCi9LnkREtYrJpDXufep_EYoRwDSJL_2z20moYMuMHy0QCg"
    )

    mocked_connection.update_authorisation_using_refresh_token(timeout=10)

    mocked_rest_call.assert_called_with(
        mocked_connection._authorisation_server_refresh_api_call,
        json={},
        headers={
            "Authorization": (
                "Bearer "
                + "eyJhbGciOiJFZERTQSIsImtpZCI6InlYaG5lSUtxUEV5Uk"
                + "lSLXVyMHdGZUZzLTZ2VS01amJEY18wUFN0X2Etc1UiLCJ0eXAiOiJKV1QifQ"
                + ".eyJhdWQiOiJodHRwczovL3FpbGltYW5qYXJvZGV2LmRkbnMubmV0OjgwODAvYXBpL3YxIiwiZXhwIjoxNjg0NDk"
                + "yMDcxLCJpYXQiOjE2ODQ0MDU2NzEsImlzcyI6Imh0dHBzOi8vcWlsaW1hbmphcm9kZXYuZGRucy5uZXQ6ODA4MC8i"
                + "LCJ0eXBlIjoicmVmcmVzaCIsInVzZXJfaWQiOjMsInVzZXJfcm9sZSI6ImFkbWluIn0"
                + ".4oSyRW9Ia7C-50x2yZxQAEXDZp-TLkFkPOtHBR4cCi9LnkREtYrJpDXufep_EYoRwDSJL_2z20moYMuMHy0QCg"
            ),
            "X-Client-Version": __version__,
        },
        timeout=10,
    )
    assert (
        mocked_connection._authorisation_access_token == web_responses.auth.raw_retrieve_response.json()["accessToken"]
    ), "Value of saved access token does not coincide with the one provided in response."


def test_refresh_token_if_unauthorised_when_ok():
    """Tests the refresh_token_if_unauthorised with a valid response"""

    func = MagicMock()
    connection = MagicMock()

    refresh_token_if_unauthorised(func=func)(self=connection)

    func.assert_called_once_with(connection)
    connection.update_authorisation_using_refresh_token.assert_not_called()


def test_refresh_token_if_unauthorised_when_unauthorised():
    """Tests the refresh_token_if_unauthorised with a valid response"""

    func = MagicMock()
    func.side_effect = [HTTPError(response=web_responses.raw.response_401), func.DEFAULT]
    connection = MagicMock()

    refresh_token_if_unauthorised(func=func)(self=connection)

    assert func.call_count == 2
    connection.update_authorisation_using_refresh_token.assert_called_once_with()


def test_refresh_token_if_unauthorised_when_other_error():
    """Tests the refresh_token_if_unauthorised with a valid response"""

    func = MagicMock()
    func.side_effect = HTTPError(response=web_responses.raw.response_500)
    connection = MagicMock()

    with pytest.raises(HTTPError):
        refresh_token_if_unauthorised(func=func)(self=connection)

    func.assert_called_once_with(connection)
    connection.update_authorisation_using_refresh_token.assert_not_called()


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_update_device_status(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """Asserts correct put function is called for updating a device status"""

    _DEVICE_ID = 1
    _NEW_STATUS = "new_status"

    mocked_connection.update_device_status(device_id=1, status=_NEW_STATUS)
    mocked_rest_call.assert_called_with(mocked_connection, path=f"/devices/{_DEVICE_ID}", data={"status": _NEW_STATUS})


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_update_device_availability(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """Asserts correct put function is called for updating a device status"""

    _DEVICE_ID = 1
    _NEW_AVAILABILITY = "new_status"

    mocked_connection.update_device_availability(device_id=1, availability=_NEW_AVAILABILITY)
    mocked_rest_call.assert_called_with(
        mocked_connection, path=f"/devices/{_DEVICE_ID}", data={"availability": _NEW_AVAILABILITY}
    )


@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_send_message(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """Asserts correct put function is called for updating a device status"""

    _CHANNEL_ID = 1
    _MESSAGE = {"block": "message"}

    mocked_connection.send_message(channel_id=_CHANNEL_ID, message=_MESSAGE)
    mocked_rest_call.assert_called_with(mocked_connection, path=f"/messages?channel={_CHANNEL_ID}", data=_MESSAGE)


@patch("qiboconnection.connection.Connection.send_post_file_auth_remote_api_call", autospec=True)
def test_send_file(mocked_rest_call: MagicMock, mocked_connection: Connection):
    """Asserts correct put function is called for updating a device status"""

    _CHANNEL_ID = 1
    _FILE = io.TextIOWrapper(buffer=io.BytesIO())
    _FILENAME = "filename"
    _TIMEOUT = 10

    mocked_connection.send_file(channel_id=_CHANNEL_ID, file=_FILE, filename=_FILENAME, timeout=_TIMEOUT)
    mocked_rest_call.assert_called_with(
        mocked_connection, path=f"/files?channel={_CHANNEL_ID}", file=_FILE, filename=_FILENAME, timeout=_TIMEOUT
    )



""" Tests methods for job result """
import numpy as np

from qiboconnection.models.job_result import JobResult
from qiboconnection.typings.enums import JobType


def test_job_result_creation():
    """Test job result creation"""
    job_result = JobResult(
        job_id=1,
        http_response="gASVsAAAAAAAAACMFW51bXB5LmNvcmUubXVsdGlhcnJheZSMDF9yZWNvbnN0cnVjdJSTlIwFbnVtcHmUjAduZGFycmF5lJOUSwCFlEMBYpSHlFKUKEsBSwWFlGgDjAVkdHlwZZSTlIwCZjiUiYiHlFKUKEsDjAE8lE5OTkr_____Sv____9LAHSUYolDKAAAAAAAAPA_AAAAAAAA8D8AAAAAAADwPwAAAAAAAPA_AAAAAAAA8D-UdJRiLg==",
        job_type=JobType.CIRCUIT,
    )

    assert isinstance(job_result, JobResult)
    assert job_result.job_id == 1
    assert (
        job_result.http_response
        == "gASVsAAAAAAAAACMFW51bXB5LmNvcmUubXVsdGlhcnJheZSMDF9yZWNvbnN0cnVjdJSTlIwFbnVtcHmUjAduZGFycmF5lJOUSwCFlEMBYpSHlFKUKEsBSwWFlGgDjAVkdHlwZZSTlIwCZjiUiYiHlFKUKEsDjAE8lE5OTkr_____Sv____9LAHSUYolDKAAAAAAAAPA_AAAAAAAA8D8AAAAAAADwPwAAAAAAAPA_AAAAAAAA8D-UdJRiLg=="
    )
    assert (job_result.data == np.array([1.0, 1.0, 1.0, 1.0, 1.0])).all()


def test_job_result_program_raises_error():
    """Test we are rising exceptions to inform correctly that PROGRAMS are not currently supported."""

    job_result = JobResult(job_id=0, http_response="WzAuMSwgMC4xLCAwLjEsIDAuMSwgMC4xXQ==", job_type="program")
    assert isinstance(job_result.data, str)



""" Tests methods for LivePlot """

import asyncio
import datetime
import json
import threading
import time
from dataclasses import asdict
from queue import Queue
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from qiboconnection.models.live_plot import WEBSOCKET_CONNECTION_LIFETIME, LivePlot
from qiboconnection.typings.live_plot import (
    LivePlotAxis,
    LivePlotLabels,
    LivePlotPacket,
    LivePlotPoints,
    LivePlotType,
    _ensure_packet_types,
)
from qiboconnection.typings.responses import PlottingResponse

from .data import heatmap_unit_plot_points, unit_plot_point
from .utils import get_current_event_loop_or_create


@pytest.fixture(name="live_plot_type")
def fixture_plot_type():
    """Returns lines plot type"""
    return LivePlotType.LINES


@pytest.fixture(name="live_plot_labels")
def fixture_plot_labels():
    """Returns a valid labels instance"""
    return LivePlotLabels()


@pytest.fixture(name="live_plot_axis")
def fixture_plot_axis():
    """Returns a valid axis instance"""
    return LivePlotAxis()


def test_ensure_packet_types_with_bad_list():
    """Test that the `_ensure_packet_types` function rises its error when a bad list is provided."""
    not_packet_list = [1, "a", None]
    with pytest.raises(ValueError):
        _ensure_packet_types(not_packet_list)


def test_ensure_packet_compatibility(
    live_plot_type: LivePlotType, live_plot_labels: LivePlotLabels, live_plot_axis: LivePlotAxis
):
    """Test that the `_ensure_packet_types` works as intended in the nominal case."""

    live_plot_packet_a = LivePlotPacket.build_packet(
        plot_id=1, plot_type=live_plot_type, labels=live_plot_labels, axis=live_plot_axis, x=0, y=0, z=None
    )
    live_plot_packet_b = LivePlotPacket.build_packet(
        plot_id=1, plot_type=live_plot_type, labels=live_plot_labels, axis=live_plot_axis, x=1, y=1, z=None
    )
    agg_live_packet = LivePlotPacket.agglutinate(packets=[live_plot_packet_a, live_plot_packet_b])

    assert agg_live_packet.data.x == [0, 1]
    assert agg_live_packet.data.y == [0, 1]


def test_ensure_packet_compatibility_with_bad_packet_list(
    live_plot_type: LivePlotType, live_plot_labels: LivePlotLabels, live_plot_axis: LivePlotAxis
):
    """Test that the `_ensure_packet_types` function rises its error when a bad list is provided."""

    live_plot_packet_a = LivePlotPacket.build_packet(
        plot_id=1, plot_type=live_plot_type, labels=live_plot_labels, axis=live_plot_axis, x=0, y=0, z=None
    )
    live_plot_packet_b = LivePlotPacket.build_packet(
        plot_id=2, plot_type=live_plot_type, labels=live_plot_labels, axis=live_plot_axis, x=1, y=1, z=None
    )

    with pytest.raises(ValueError):
        LivePlotPacket.agglutinate(packets=[live_plot_packet_a, live_plot_packet_b])


def test_live_plot(live_plot_type: LivePlotType, live_plot_labels: LivePlotLabels, live_plot_axis: LivePlotAxis):
    """test live plot"""

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    assert live_plot.plot_id == plot_id
    assert live_plot.plot_type == live_plot_type
    assert live_plot.labels == live_plot_labels
    assert live_plot.axis == live_plot_axis


@pytest.fixture(name="live_plot_points")
def fixture_live_plot_points() -> LivePlotPoints:
    """Returns a Valid points instance"""
    return LivePlotPoints(x=unit_plot_point[0]["x"], y=unit_plot_point[0]["y"])


def test_live_plot_points_constructor(live_plot_points: LivePlotPoints):
    """test live plot creation with points"""

    assert isinstance(live_plot_points, LivePlotPoints)
    assert live_plot_points.x == unit_plot_point[0]["x"]
    assert live_plot_points.y == unit_plot_point[0]["y"]
    assert live_plot_points.to_scatter() == [unit_plot_point[0]]


def test_live_plot_points_equality():
    """test live plots `__eq__`"""

    live_plot_points_1 = LivePlotPoints(x=0, y=0)
    live_plot_points_2 = LivePlotPoints(x=0, y=0)
    assert live_plot_points_1 == live_plot_points_2

    live_plot_points_3 = LivePlotPoints(x=0, y=0, z=0)
    assert live_plot_points_1 != live_plot_points_3

    live_plot_points_4 = "ThisIsNotALivePointsObject"
    assert live_plot_points_1 != live_plot_points_4


def test_live_plot_points_raises_value_error_for_mixed_array_like_with_number_like_input():
    """test live plot fails when mixing points and arrays"""

    with pytest.raises(ValueError) as e_info:
        _ = LivePlotPoints(x=0, y=[0])

    assert e_info.value.args[0] == "Arguments provided must be of the same type: floats or lists"


@pytest.fixture(name="heatmap_plot_type")
def fixture_heatmap_plot_type() -> LivePlotType:
    """Returns heatmap type"""
    return LivePlotType.HEATMAP


@pytest.fixture(name="heatmap_plot_axis")
def fixture_heatmap_plot_axis() -> LivePlotAxis:
    """Returns axis instance valid for heatmap"""
    return LivePlotAxis(
        x_axis=[point["x"] for point in heatmap_unit_plot_points],
        y_axis=[point["y"] for point in heatmap_unit_plot_points],
    )


@pytest.fixture(name="heatmap_plot_points")
def fixture_heatmap_plot_points() -> LivePlotPoints:
    """Returns plot points valid for heatmap"""
    assert all(None not in point.values() for point in heatmap_unit_plot_points)
    x = [int(point["x"]) for point in heatmap_unit_plot_points]  # type: ignore
    y = [int(point["y"]) for point in heatmap_unit_plot_points]  # type: ignore
    z = [int(point["z"]) for point in heatmap_unit_plot_points]  # type: ignore
    idx = [int(point["idx"]) for point in heatmap_unit_plot_points]  # type: ignore
    idy = [int(point["idy"]) for point in heatmap_unit_plot_points]  # type: ignore

    return LivePlotPoints(x=x, y=y, z=z, idx=idx, idy=idy)


def test_heatmap_data_packet(
    heatmap_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    heatmap_plot_axis: LivePlotAxis,
    heatmap_plot_points: LivePlotPoints,
):
    """Tests heatmap data packet"""

    plot_id = 1
    live_plot_packet = LivePlotPacket.build_packet(
        plot_id=plot_id,
        plot_type=heatmap_plot_type,
        labels=live_plot_labels,
        axis=heatmap_plot_axis,
        x=[point["x"] for point in heatmap_unit_plot_points],
        y=[point["y"] for point in heatmap_unit_plot_points],
        z=[point["z"] for point in heatmap_unit_plot_points],
    )

    assert live_plot_packet.plot_id == plot_id
    assert live_plot_packet.plot_type == heatmap_plot_type
    assert live_plot_packet.labels == live_plot_labels
    assert live_plot_packet.axis == heatmap_plot_axis
    assert live_plot_packet.data == heatmap_plot_points
    assert live_plot_packet == LivePlotPacket(
        plot_id=plot_id,
        plot_type=heatmap_plot_type,
        labels=live_plot_labels,
        axis=heatmap_plot_axis,
        data=heatmap_plot_points,
    )
    assert live_plot_packet.to_json() == json.dumps(
        {
            "plot_id": plot_id,
            "plot_type": heatmap_plot_type.value,
            "labels": asdict(live_plot_labels),
            "data": heatmap_plot_points.to_scatter(),
        }
    )


def test_heatmap_data_packet_with_numpy_arrays(
    heatmap_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    heatmap_plot_axis: LivePlotAxis,
    heatmap_plot_points: LivePlotPoints,
):
    """Tests heatmap data packet with arrays"""
    plot_id = 1
    live_plot_packet = LivePlotPacket.build_packet(
        plot_id=plot_id,
        plot_type=heatmap_plot_type,
        labels=live_plot_labels,
        axis=heatmap_plot_axis,
        x=np.array([point["x"] for point in heatmap_unit_plot_points]),
        y=np.array([point["y"] for point in heatmap_unit_plot_points]),
        z=np.array([point["z"] for point in heatmap_unit_plot_points]),
    )

    assert live_plot_packet.to_json() == json.dumps(
        {
            "plot_id": plot_id,
            "plot_type": heatmap_plot_type.value,
            "labels": asdict(live_plot_labels),
            "data": heatmap_plot_points.to_scatter(),
        }
    )


def test_heatmap_data_packet_parsing_with_point_single_values(
    heatmap_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    heatmap_plot_axis: LivePlotAxis,
    heatmap_plot_points: LivePlotPoints,
):
    """Tests heatmap data packet with single points"""
    plot_id = 1
    live_plot_packets = [
        LivePlotPacket.build_packet(
            plot_id=plot_id,
            plot_type=heatmap_plot_type,
            labels=live_plot_labels,
            axis=heatmap_plot_axis,
            x=point["x"],
            y=point["y"],
            z=point["z"],
        )
        for point in heatmap_unit_plot_points
    ]

    for index, live_plot_packet in enumerate(live_plot_packets):
        assert live_plot_packet.to_json() == json.dumps(
            {
                "plot_id": plot_id,
                "plot_type": heatmap_plot_type.value,
                "labels": asdict(live_plot_labels),
                "data": LivePlotPoints(
                    x=heatmap_plot_points.x[index],
                    y=heatmap_plot_points.y[index],
                    z=heatmap_plot_points.z[index],
                    idx=heatmap_plot_points._idx[index],  # type: ignore
                    idy=heatmap_plot_points._idy[index],  # type: ignore
                ).to_scatter(),
            }
        )


def test_plotting_response_constructor():
    """tests plotting response"""
    plotting_response = PlottingResponse(websocket_url="server/demo-url", plot_id=1)

    assert isinstance(plotting_response, PlottingResponse)
    assert plotting_response.plot_id == 1
    assert plotting_response.websocket_url == "server/demo-url"


def test_plotting_response_from_response():
    """tests plotting response `from_response`"""
    plotting_response_input = {"websocket_url": "server/demo-url", "plot_id": "1"}
    plotting_response = PlottingResponse.from_response(**plotting_response_input)

    assert isinstance(plotting_response, PlottingResponse)
    assert plotting_response.plot_id == int(plotting_response_input["plot_id"])
    assert plotting_response.websocket_url == str(plotting_response_input["websocket_url"])


def test_plotting_response_to_dict():
    """tests plotting response `to_dict`"""
    plotting_response_dict = PlottingResponse(websocket_url="server/demo-url", plot_id=1).to_dict()

    assert isinstance(plotting_response_dict, dict)
    assert plotting_response_dict["plot_id"] == 1
    assert plotting_response_dict["websocket_url"] == "server/demo-url"


@patch("websockets.connect", autospec=True)
def test_start_up(
    mocked_websockets_connect: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Tests start-up method for the initialisation of a LivePlot"""

    loop = get_current_event_loop_or_create()

    mocked_connection_future = loop.create_future()
    mocked_connection_future.set_result(MagicMock())
    mocked_websockets_connect.return_value = mocked_connection_future

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    asyncio.run(live_plot.start_up())

    mocked_websockets_connect.assert_called_with(websocket_url)
    assert isinstance(live_plot._send_queue_thread, threading.Thread)
    assert isinstance(live_plot._send_queue, Queue)


@patch("websockets.connect", autospec=True)
def test_open_connection_with_failing_connection(
    mocked_websockets_connect: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Test _open_connection and _close_connection work as intended"""

    mocked_connection_connect_return = MagicMock(side_effect=Exception("Test exception"))
    mocked_websockets_connect.return_value = mocked_connection_connect_return

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    with pytest.raises(ValueError):
        asyncio.run(live_plot._open_connection())
    assert live_plot._connection is None
    assert live_plot._connection_started_at is None


@patch("websockets.connect", autospec=True)
def test_open_and_close_connection(
    mocked_websockets_connect: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Test _open_connection and _close_connection work as intended"""

    loop = get_current_event_loop_or_create()

    mocked_connection_close = MagicMock()

    mocked_connection_close_future = loop.create_future()
    mocked_connection_close_future.set_result(mocked_connection_close)

    mocked_connection = MagicMock()
    mocked_connection.open = True
    mocked_connection.close.return_value = mocked_connection_close_future

    mocked_connection_connect_future = loop.create_future()
    mocked_connection_connect_future.set_result(mocked_connection)

    mocked_websockets_connect.return_value = mocked_connection_connect_future

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    asyncio.run(live_plot._open_connection())
    mocked_websockets_connect.assert_called_with("test_url")
    assert live_plot._connection is not None
    assert isinstance(live_plot._connection_started_at, datetime.datetime)

    asyncio.run(live_plot._close_connection())
    mocked_connection.close.assert_called()
    assert live_plot._connection is None
    assert live_plot._connection_started_at is None


@patch("qiboconnection.models.live_plot.LivePlot._open_connection", autospec=True)
@patch("qiboconnection.models.live_plot.LivePlot._close_connection", autospec=True)
def test_reset_connection(
    mocked_open_connection: MagicMock,
    mocked_close_connection: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Tests _reset_connection method closes and opens a connection"""

    loop = get_current_event_loop_or_create()

    mocked_open_connection_return = MagicMock()
    mocked_open_connection_future = loop.create_future()
    mocked_open_connection_future.set_result(mocked_open_connection_return)
    mocked_open_connection.return_value = mocked_open_connection_future

    mocked_close_connection_return = MagicMock()
    mocked_close_connection_future = loop.create_future()
    mocked_close_connection_future.set_result(mocked_close_connection_return)
    mocked_close_connection.return_value = mocked_close_connection_future

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    asyncio.run(live_plot._reset_connection())
    mocked_close_connection.assert_called()
    mocked_open_connection.assert_called()


@patch("qiboconnection.models.live_plot.LivePlot._close_connection")
@patch("websockets.connect", autospec=True)
def test_send_data(
    mocked_websockets_connect: MagicMock,
    _mocked_close_connection: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Tests send_data method in the"""

    loop = get_current_event_loop_or_create()

    mocked_connection_send_future = loop.create_future()
    mocked_connection_send_future.set_result("OK")
    mocked_connection = MagicMock()
    mocked_connection.open = True
    mocked_connection.send.return_value = mocked_connection_send_future

    mocked_connection_connect_future = loop.create_future()
    mocked_connection_connect_future.set_result(mocked_connection)
    mocked_websockets_connect.return_value = mocked_connection_connect_future

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    asyncio.run(live_plot.start_up())

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=LivePlotType.LINES,
        x=1,
        y=2,
        z=None,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )
    live_plot._connection_started_at = datetime.datetime.now()
    asyncio.run(live_plot.send_data(data=data_packet))

    time.sleep(1)
    mocked_connection.send.assert_called_with(data_packet.to_json())


@patch("qiboconnection.models.live_plot.LivePlot._reset_connection", autospec=True)
@patch("websockets.connect", autospec=True)
def test_send_data_with_too_log_lived_connection(
    mocked_websockets_connect: MagicMock,
    mocked_reset_connection: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Tests start-up method for the initialisation of a LivePlot"""
    # pylint: disable=too-many-locals

    loop = get_current_event_loop_or_create()

    mocked_connection_send_future = loop.create_future()
    mocked_connection_send_future.set_result("OK")
    mocked_connection_close_future = loop.create_future()
    mocked_connection_close_future.set_result("OK")
    mocked_connection = MagicMock()
    mocked_connection.open = True
    mocked_connection.send.return_value = mocked_connection_send_future
    mocked_connection.close.return_value = mocked_connection_close_future

    mocked_connection_connect_future = loop.create_future()
    mocked_connection_connect_future.set_result(mocked_connection)
    mocked_websockets_connect.return_value = mocked_connection_connect_future

    mocked_reset_connection_return = MagicMock()
    mocked_reset_connection_future = loop.create_future()
    mocked_reset_connection_future.set_result(mocked_reset_connection_return)
    mocked_reset_connection.return_value = mocked_reset_connection_future

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    asyncio.run(live_plot.start_up())

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=LivePlotType.LINES,
        x=1,
        y=2,
        z=None,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )
    live_plot._connection_started_at = datetime.datetime.now() - datetime.timedelta(
        seconds=WEBSOCKET_CONNECTION_LIFETIME + 1
    )
    asyncio.run(live_plot.send_data(data=data_packet))

    time.sleep(1)
    mocked_reset_connection.assert_called()
    mocked_connection.send.assert_called_with(data_packet.to_json())


@patch("qiboconnection.models.live_plot.LivePlot._reset_connection", autospec=True)
@patch("websockets.connect", autospec=True)
def test_send_data_with_closed_connection(
    mocked_websockets_connect: MagicMock,
    mocked_reset_connection: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Tests start-up method for the initialisation of a LivePlot"""

    loop = get_current_event_loop_or_create()

    mocked_connection_send_future = loop.create_future()
    mocked_connection_send_future.set_result("OK")
    mocked_connection = MagicMock()
    mocked_connection.open = False
    mocked_connection.send.return_value = mocked_connection_send_future

    mocked_connection_connect_future = loop.create_future()
    mocked_connection_connect_future.set_result(mocked_connection)
    mocked_websockets_connect.return_value = mocked_connection_connect_future

    mocked_reset_connection_return = MagicMock()
    mocked_reset_connection_future = loop.create_future()
    mocked_reset_connection_future.set_result(mocked_reset_connection_return)
    mocked_reset_connection.return_value = mocked_reset_connection_future

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    asyncio.run(live_plot.start_up())

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=LivePlotType.LINES,
        x=1,
        y=2,
        z=None,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )
    live_plot._connection_started_at = datetime.datetime.now()
    asyncio.run(live_plot.send_data(data=data_packet))

    time.sleep(1)
    mocked_reset_connection.assert_called()


@patch("qiboconnection.models.live_plot.LivePlot._setup_queue", autospec=True)
@patch("websockets.connect", autospec=True)
def test_send_data_with_missbehaving_queue(
    mocked_websockets_connect: MagicMock,
    mocked_setup_queue: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Tests start-up method for the initialisation of a LivePlot"""

    loop = get_current_event_loop_or_create()

    mocked_connection_send_future = loop.create_future()
    mocked_connection_send_future.set_result("OK")
    mocked_connection = MagicMock()
    mocked_connection.open = False
    mocked_connection.send.return_value = mocked_connection_send_future

    mocked_connection_connect_future = loop.create_future()
    mocked_connection_connect_future.set_result(mocked_connection)
    mocked_websockets_connect.return_value = mocked_connection_connect_future

    mocked_setup_queue.return_value = None

    plot_id = 1
    websocket_url = "test_url"
    live_plot = LivePlot(
        plot_id=plot_id,
        plot_type=live_plot_type,
        websocket_url=websocket_url,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=LivePlotType.LINES,
        x=1,
        y=2,
        z=None,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )
    live_plot._connection_started_at = datetime.datetime.now()
    with pytest.raises(ValueError):
        asyncio.run(live_plot.send_data(data=data_packet))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" SavedExperimentResponse Typing"""
# pylint: disable=too-many-instance-attributes
from dataclasses import dataclass
from datetime import datetime


@dataclass
class SavedExperimentResponse:
    """Encodes the expected response of the server when requested for a specific SavedExperiment"""

    name: str
    experiment: str
    device_id: int
    user_id: int
    description: str
    qililab_version: str
    results: str
    saved_experiment_id: int
    created_at: datetime



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" SavedExperimentListing class """

from dataclasses import asdict, dataclass, field
from typing import List

import pandas as pd

from qiboconnection.typings.responses import SavedExperimentListingItemResponse

from .saved_experiment_listing_item import SavedExperimentListingItem


@dataclass
class SavedExperimentListing:
    """SavedExperimentListing representation"""

    items: List[SavedExperimentListingItem]
    _dataframe: pd.DataFrame = field(init=False)

    def __post_init__(self):
        self._dataframe = self._build_dataframe()
        self._coerce_dataframe_column_types()

    def _build_dataframe(self):
        """Builds the dataframe from the info in each listing item"""
        return pd.DataFrame((asdict(item) for item in self.items))

    def _coerce_dataframe_column_types(self):
        """Parses each column that can potentially have its type misinterpretated to the one we would expect"""
        self._dataframe["created_at"] = pd.to_datetime(self._dataframe["created_at"])

    @classmethod
    def from_response(cls, response_list: List[SavedExperimentListingItemResponse]):
        """Constructor for SavedExperimentListing that takes in a list of SavedExperimentListingItemResponse"""
        return cls(items=[SavedExperimentListingItem.from_response(response=response) for response in response_list])

    @property
    def dataframe(self):
        """Returns the dataframe shape of the results. Might be useful for performing queries over listed data."""
        return self._dataframe



""" Tests methods for User """

from qiboconnection.models.user import User


def test_user_constructor():
    """Test User class constructor."""
    user = User(user_id=1, username="test-user", api_key="000-3333")
    assert isinstance(user, User)
    print(user.__dict__)
    assert user.user_id == 1
    assert user.username == "test-user"
    assert user.api_key == "000-3333"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" JobListing class """

from dataclasses import asdict, dataclass, field
from typing import List

import pandas as pd

from qiboconnection.typings.responses import JobListingItemResponse

from .job_listing_item import JobListingItem


@dataclass
class JobListing:
    """JobListing representation"""

    items: List[JobListingItem]
    _dataframe: pd.DataFrame = field(init=False)

    def __post_init__(self):
        self._dataframe = self._build_dataframe()

    def _build_dataframe(self):
        """Builds the dataframe from the info in each listing item"""
        return pd.DataFrame((asdict(item) for item in self.items))

    @classmethod
    def from_response(cls, response_list: List[JobListingItemResponse]):
        """Constructor for JobListing that takes in a list of JobResponse"""
        return cls(items=[JobListingItem.from_response(response=response) for response in response_list])

    @property
    def dataframe(self):
        """Returns the dataframe shape of the results. Might be useful for performing queries over listed data."""
        return self._dataframe



""" Tests methods for Devices """

from typing import List

import pytest

from qiboconnection.models.devices import SimulatorDevice
from qiboconnection.models.devices.device import Device
from qiboconnection.models.devices.devices import Devices
from qiboconnection.typings.devices import SimulatorDeviceInput

from .data import simulator_device_inputs


def test_devices_constructor():
    """Test Devices class constructor"""
    devices = Devices()
    assert isinstance(devices, Devices)


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_devices_constructor_with_single_device(simulator_device_input: SimulatorDeviceInput):
    """Test Devices class constructor providing a single device"""
    simulator_device: Device = SimulatorDevice(device_input=simulator_device_input)
    devices = Devices(device=simulator_device)
    assert isinstance(devices, Devices)


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_devices_constructor_with_device_list(simulator_device_input: SimulatorDeviceInput):
    """Test Devices class constructor providing a list of devices"""
    simulator_device_list: List[Device] = [SimulatorDevice(device_input=simulator_device_input)]
    devices = Devices(device=simulator_device_list)
    assert isinstance(devices, Devices)


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_devices_json(simulator_device_input: SimulatorDeviceInput):
    """Test Devices().toJSON() method"""
    simulator_device = SimulatorDevice(device_input=simulator_device_input)
    devices = Devices(device=simulator_device)
    assert devices.toJSON() == f"{simulator_device.toJSON()}\n"


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_devices_str(simulator_device_input: SimulatorDeviceInput):
    """Test Devices().__str__() method"""
    simulator_device = SimulatorDevice(device_input=simulator_device_input)
    devices = Devices(device=simulator_device)
    assert str(devices) == f"<Devices[1]:\n{simulator_device.toJSON()}\n"


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_devices_update_device(simulator_device_input: SimulatorDeviceInput):
    """Test Devices().add_or_update() method with a not-preexisting device (add) and with a preexisting device
    (update)"""
    simulator_device = SimulatorDevice(device_input=simulator_device_input)
    devices = Devices()
    # ADDS
    devices.add_or_update(simulator_device)
    assert devices.to_dict() == [dict(simulator_device.__dict__.items())]
    # UPDATES
    devices.add_or_update(simulator_device)
    assert devices.to_dict() == [dict(simulator_device.__dict__.items())]


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_devices_select_device(simulator_device_input: SimulatorDeviceInput):
    """Test Devices().select_device() method"""
    simulator_device = SimulatorDevice(device_input=simulator_device_input)
    devices = Devices(simulator_device)
    selected_device = devices.select_device(device_id=1)
    assert selected_device.toJSON() == simulator_device.toJSON()


def test_devices_select_device_raises_value_error_for_nonexistent_devices():
    """Test Devices().select_device() method providing it an id of a non-existing device."""
    devices = Devices()
    with pytest.raises(ValueError) as e_info:
        _ = devices.select_device(device_id=1)
    assert e_info.value.args[0] == "Device not found"


@pytest.mark.parametrize("simulator_device_input", simulator_device_inputs)
def test_devices_select_device_raises_value_error_for_duplicated(simulator_device_input: SimulatorDeviceInput):
    """Test Devices.select_device() method providing it an id of a duplicated device."""
    simulator_device = SimulatorDevice(device_input=simulator_device_input)
    devices = Devices([simulator_device, simulator_device])
    with pytest.raises(ValueError) as e_info:
        _ = devices.select_device(device_id=1)
    assert e_info.value.args[0] == "Device duplicated with same id:1"



""" Test methods for SavedExperiment classes"""

import datetime

from qiboconnection.models.saved_experiment import SavedExperiment
from qiboconnection.typings.requests import SavedExperimentRequest
from qiboconnection.typings.responses import SavedExperimentResponse


def test_saved_experiment_creation():
    """Tests SavedExperiment creation"""
    saved_experiment = SavedExperiment(
        name="name",
        user_id=1,
        device_id=1,
        description="description",
        experiment={"a": 0},
        results={"a": 0},
        qililab_version="0.0.0",
        id=0,
        created_at=datetime.datetime.now(),
    )

    assert isinstance(saved_experiment, SavedExperiment)


def test_saved_experiment_creation_from_response():
    """Tests SavedExperiment creation from response"""
    saved_experiment_response = SavedExperimentResponse(
        name="name",
        user_id=1,
        device_id=1,
        description="description",
        experiment="eyJhIjogMH0=",
        results="eyJhIjogMH0=",
        qililab_version="0.0.0",
        saved_experiment_id=0,
        created_at=datetime.datetime.now(),
    )

    saved_experiment = SavedExperiment.from_response(response=saved_experiment_response)

    assert isinstance(saved_experiment, SavedExperiment)
    assert saved_experiment.experiment == {"a": 0}, "Decoded experiment does not coincide with expected"
    assert saved_experiment.results == {"a": 0}, "Decoded results do not coincide with expected"


def test_saved_experiment_request():
    """Tests SavedExperiment request builder method"""
    saved_experiment = SavedExperiment(
        name="name",
        user_id=1,
        device_id=1,
        description="description",
        experiment={"a": 0},
        results={"a": 0},
        qililab_version="0.0.0",
        id=0,
        created_at=datetime.datetime.now(),
    )

    saved_experiment_request = saved_experiment.saved_experiment_request(favourite=False)

    assert isinstance(saved_experiment_request, SavedExperimentRequest)
    assert saved_experiment_request.experiment == "eyJhIjogMH0=", "Decoded experiment does not coincide with expected"
    assert saved_experiment_request.results == "eyJhIjogMH0=", "Decoded results do not coincide with expected"



from unittest.mock import MagicMock, patch

import pytest

from qiboconnection.typings.job_data import JobData
from qiboconnection.typings.responses import JobListingItemResponse


def test_JobData_typing():
    """Test JobData typing, this is what is returned to the user with get_job()."""

    job_full_data = JobData(
        queue_position=32,
        status="89",
        user_id=None,
        device_id=3,
        job_id=4,
        job_type="jaiof",
        number_shots=84,
        description="<qibo.models.circuit.Circuit at 0x7fb8c2127650>",
        result={},
    )
    assert isinstance(job_full_data, JobData)


def test_JobData_typing_result_raises_value_error():
    """Test JobData typing raises error if results are not the correct type"""

    with pytest.raises(ValueError) as ex:
        _ = JobData(
            queue_position=32,
            status="completed",
            user_id=None,
            device_id=3,
            job_id=4,
            job_type="jaiof",
            number_shots=84,
            description="<qibo.models.circuit.Circuit at 0x7fb8c2127650>",
            result="Not a dict or list",
        )

    assert ex.match("Job result needs to be a dict, a list or a None!")


@patch("qiboconnection.typings.job_data.deserialize_job_description", autospec=True)
def test_JobData_typing_descripton_raises_value_error(mocked_deserialize_job_description: MagicMock):
    """Test JobData typing, this is what is returned to the user with get_job()."""

    mocked_deserialize_job_description.return_value = "A bad description"

    with pytest.raises(ValueError) as ex:
        _ = JobData(
            queue_position=32,
            status="invent",
            user_id=None,
            device_id=3,
            job_id=4,
            job_type="experiment",
            number_shots=84,
            description="",
            result="",
        )

    assert ex.match("Job description needs to be a Qibo Circuit, a dict, a list or a None!")


def test_job_listing_item_response_typing():
    """Test JobListingItemResponse instantiation"""

    listing_job_response = JobListingItemResponse(
        status="89", user_id=None, device_id=3, job_type="jaiof", number_shots=84, id=3
    )

    assert isinstance(listing_job_response, JobListingItemResponse)



""" Test methods for SavedExperiment typing classes"""

import datetime

from qiboconnection.typings.requests import SavedExperimentRequest
from qiboconnection.typings.responses import SavedExperimentListingItemResponse, SavedExperimentResponse


def test_saved_experiment_request_creation():
    """Tests SavedExperimentRequest creation"""
    saved_experiment_request = SavedExperimentRequest(
        name="name",
        experiment="eyJhIjogMH0=",
        results="eyJhIjogMH0=",
        device_id=1,
        user_id=1,
        description="description",
        qililab_version="0.0.0",
        favourite=False,
    )

    assert isinstance(saved_experiment_request, SavedExperimentRequest)


def test_saved_experiment_response_creation():
    """Tests SavedExperimentResponse creation"""
    saved_experiment_response = SavedExperimentResponse(
        name="name",
        experiment="eyJhIjogMH0=",
        results="eyJhIjogMH0=",
        device_id=1,
        user_id=1,
        description="description",
        qililab_version="0.0.0",
        saved_experiment_id=0,
        created_at=datetime.datetime.now(),
    )

    assert isinstance(saved_experiment_response, SavedExperimentResponse)


def test_saved_experiment_listing_item_response_creation():
    """Tests SavedExperimentListingItemResponse creation"""
    saved_experiment_listing_item_response = SavedExperimentListingItemResponse(
        name="name",
        experiment="eyJhIjogMH0=",
        device_id=1,
        user_id=1,
        description="description",
        qililab_version="0.0.0",
        id=0,
        created_at=datetime.datetime.now(),
    )

    assert isinstance(saved_experiment_listing_item_response, SavedExperimentListingItemResponse)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Simulator Device Characteristics Input Typing """
from dataclasses import dataclass
from typing import Literal

from qiboconnection.typings.enums import DeviceType


@dataclass(kw_only=True)
class SimulatorDeviceCharacteristicsInput:
    """Simulator Device Characteristics Input

    Attributes:
        type (str): device type, "simulator"
        cpu (str): device cpu
        gpu (str): device gpu
        os (str): device os
        kernel (str): device kernel
        ram (str): device ram

    """

    type: Literal[DeviceType.SIMULATOR, "simulator"]
    cpu: str | None = None
    gpu: str | None = None
    os: str | None = None  # pylint: disable=invalid-name
    kernel: str | None = None
    ram: str | None = None



import pandas as pd

from qiboconnection.models.job_listing import JobListing, JobListingItem
from qiboconnection.typings.responses import JobListingItemResponse


def test_job_listing_item_creation_from_response():
    """JobListingItem cration from response"""
    listing_job_response = JobListingItemResponse(
        status="89", user_id=None, device_id=3, job_type="jaiof", number_shots=84, id=3
    )
    Job_listing_item = JobListingItem.from_response(listing_job_response)

    assert isinstance(Job_listing_item, JobListingItem)
    assert Job_listing_item.id == listing_job_response.id


def test_job_listing_creation_dataframe():
    """Ensure job are listed as dataframes"""
    listing_job_response_1 = JobListingItemResponse(
        status="89", user_id=5, device_id=3, job_type="jaiof", number_shots=84, id=3
    )
    listing_job_response_2 = JobListingItemResponse(
        status="89", user_id=5, device_id=38, job_type="jahfwsiof", number_shots=884, id=3
    )
    job_listing = JobListing.from_response([listing_job_response_1, listing_job_response_2])

    assert isinstance(job_listing, JobListing)
    assert isinstance(job_listing.dataframe, pd.DataFrame)



""" Test Error rising and utils """

import pytest
from requests.models import HTTPError, Response

from qiboconnection.errors import ConnectionException, RemoteExecutionException, custom_raise_for_status


@pytest.fixture(name="response_400")
def fixture_response_400():
    """Build the response object associated to a 400 Error Response

    Returns:
        Response: response object
    """
    response = Response()
    response.status_code = 400
    response.url = "server/api"
    response._content = b'{"DEMO": "400"}'
    response.reason = "expired"
    return response


@pytest.fixture(name="response_500")
def fixture_response_500():
    """Build the response object associated to a 500 Error Response

    Returns:
        Response: response object
    """
    response = Response()
    response.status_code = 500
    response.url = "server/api"
    response._content = b'{"DEMO": "500"}'
    response.reason = "internal server error"
    return response


def test_remote_execution_exception():
    """Test the RemoteExecutionException is raised properly when explicitly raised."""
    with pytest.raises(RemoteExecutionException) as e_info:
        raise RemoteExecutionException("DEMO", status_code=500)
    assert e_info.type == RemoteExecutionException
    assert e_info.value.args[0] == "DEMO"
    assert e_info.value.status_code == 500


def test_connection_exception():
    """Test the ConnectionException is raised properly when explicitly raised."""
    with pytest.raises(ConnectionException) as e_info:
        raise ConnectionException
    assert e_info.type == ConnectionException


def test_custom_raise_for_status_400(response_400: Response):
    """Test custom_raise_for_status() raises the proper exception when meeting a status_code 400 response."""
    with pytest.raises(HTTPError) as e_info:
        custom_raise_for_status(response_400)
    assert e_info.type == HTTPError
    assert e_info.value.response == response_400


def test_custom_raise_for_status_500(response_500: Response):
    """Test custom_raise_for_status() raises the proper exception when meeting a status_code 500 response."""
    with pytest.raises(HTTPError) as e_info:
        custom_raise_for_status(response_500)
    assert e_info.type == HTTPError
    assert e_info.value.response == response_500



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" LivePlot class """

import asyncio
import contextlib
import datetime
import os
import threading
import time
from abc import ABC
from dataclasses import asdict
from queue import Queue
from ssl import SSLError
from typing import List

import websockets
from websockets.exceptions import ConnectionClosed, WebSocketException

from qiboconnection.config import logger
from qiboconnection.typings.live_plot import LivePlotAxis, LivePlotLabels, LivePlotPacket, LivePlotType

WEBSOCKET_CONNECTION_LIFETIME = int(os.getenv("QIBOCONNECTION_WEBSOCKET_CONNECTION_LIFETIME", default="5"))
PACKET_POINT_NUMBER_LIMIT = int(os.getenv("QIBOCONNECTION_PACKET_POINT_NUMBER_LIMIT", default="1000"))


class LivePlot(ABC):  # pylint: disable=too-many-instance-attributes
    """Job class to manage the job experiment to be remotely sent"""

    def __init__(
        self, plot_id: int, plot_type: LivePlotType, websocket_url: str, labels: LivePlotLabels, axis: LivePlotAxis
    ):
        """Default constructor. _connection is left as None until it is tried to use"""
        self._plot_id: int = plot_id
        self._plot_type: LivePlotType = plot_type
        self._labels: LivePlotLabels = labels
        self._axis: LivePlotAxis = axis
        self._websocket_url: str = websocket_url
        # pylint: disable=no-member
        self._connection: websockets.WebSocketClientProtocol | None = None  # type: ignore [name-defined]
        self._connection_open = False
        self._send_queue: Queue | None = None
        self._send_queue_thread: threading.Thread | None = None
        self._connection_started_at: datetime.datetime | None = None

    async def start_up(self):
        """Sets up the queue, opens the connection and creates and starts the sending data thread"""
        self._setup_queue()
        await self._open_connection()
        self._setup_sending_thread()

    def _setup_queue(self):
        """Sets up an empty queue"""
        self._send_queue = Queue()

    def _setup_sending_thread(self):
        """Sets up and starts the sending thread"""
        self._send_queue_thread = threading.Thread(target=asyncio.run, args=(self._sending_loop(),))
        self._send_queue_thread.daemon = True
        self._send_queue_thread.start()

    async def _reset_connection_if_opened_for_too_long(self):
        """Resets the connection if it has been up for too long"""
        elapsed_time = datetime.datetime.now() - self._connection_started_at
        if elapsed_time.seconds > WEBSOCKET_CONNECTION_LIFETIME:
            await self._reset_connection()

    def _consume_and_agglutinate_all_packets_in_queue(self):
        """Consumes all packets in queue into a list and agglutinates them into a single packet"""

        packet_list: List[LivePlotPacket] = []
        while not self._send_queue.empty() and len(packet_list) < PACKET_POINT_NUMBER_LIMIT:
            packet_list.append(self._send_queue.get())
        return LivePlotPacket.agglutinate(packets=packet_list)

    @property
    def connection_open(self):
        """Return true if the connection is open"""
        return self._connection.open if self._connection else False

    def ensure_connection_open(self):
        """Raises ValueError if the connection is not open

        Raises:
           ValueError: The connection is not open
        """
        if not self.connection_open:
            raise ValueError("The connection is not open")

    async def _sending_loop(self):
        """Main loop, where we send over the socket whatever there is in the _send_queue"""
        while True:
            await self._reset_connection_if_opened_for_too_long()

            if agglutinated_packet := self._consume_and_agglutinate_all_packets_in_queue():
                agglutinated_message = agglutinated_packet.to_json()
                try:
                    self.ensure_connection_open()
                    await self._connection.send(agglutinated_message)
                except (
                    AssertionError,
                    AttributeError,
                    ValueError,
                    SSLError,
                    WebSocketException,
                    ConnectionClosed,
                ) as ex:
                    logger.debug(f"Found error {ex} ({type(ex)}) while plotting. Resetting connection")
                    await self._reset_connection()
                    await self._connection.send(agglutinated_message)
            else:
                time.sleep(0.2)

    @property
    def plot_id(self) -> int:
        """Gets the plot_id."""
        return self._plot_id

    @property
    def plot_type(self) -> LivePlotType:
        """Gets the plot_type."""
        return self._plot_type

    @property
    def labels(self) -> LivePlotLabels:
        """Gets the plot_type."""
        return self._labels

    @property
    def axis(self) -> LivePlotAxis:
        """Gets the plot_type."""
        return self._axis

    async def send_data(self, data: LivePlotPacket):
        """Sends a LivePlotPacket over the websocket connection.
        Returns:
            Length of message sent.
        """
        if self._send_queue is None:
            self._setup_queue()
        if isinstance(self._send_queue, Queue):
            self._send_queue.put(data)
            return
        raise ValueError("Queue was not properly set up")

    async def _close_connection(self):
        """
        Closes the created connection and restores the variable to None.
        """
        logger.debug("Closing connection")
        if self._connection is not None:
            with contextlib.suppress(AttributeError, ValueError, SSLError, WebSocketException, ConnectionClosed):
                await self._connection.close()
            self._connection = None
            self._connection_started_at = None

    async def _open_connection(self):
        """
        Creates connection using self._websocket_url and saves it to self._connection.
        """
        logger.debug("Opening connection")
        try:
            self._connection = await websockets.connect(self._websocket_url)  # pylint: disable=no-member
            self._connection_started_at = datetime.datetime.now()
        except Exception as ex:
            raise ValueError("Server is refusing connections") from ex

    async def _reset_connection(self):
        """
        Resets the connection.
        """
        logger.debug("Resetting connection")
        await self._close_connection()
        await self._open_connection()

    def __eq__(self, other):
        return (
            (
                other.plot_id == self.plot_id
                and other.plot_type == self.plot_type
                and asdict(other.axis) == asdict(self.axis)
                and asdict(self.labels) == asdict(self.labels)
                and other._websocket_url == self._websocket_url
            )
            if isinstance(other, LivePlot)
            else False
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Calibration Details Input Typing """
from dataclasses import dataclass


@dataclass(kw_only=True)
class CalibrationDetailsInput:
    """Calibration Details Input

    Attributes:
        elapsed_time (int | None): elapsed time
        t1 (int | None): last calibrated t1 time
        frequency (int | None): last calibrated frequency
    """

    elapsed_time: int | None = None
    t1: int | None = None  # pylint: disable=invalid-name
    frequency: int | None = None



""" Tests methods for LivePlots """

import asyncio
from unittest.mock import MagicMock, patch

import pytest

from qiboconnection.models.live_plot import LivePlot
from qiboconnection.models.live_plots import LivePlots
from qiboconnection.typings.live_plot import LivePlotAxis, LivePlotLabels, LivePlotPacket, LivePlotType

from .utils import get_current_event_loop_or_create


def ok_method(*_, **__):
    """returns `'ok'`"""
    return "OK"


@pytest.fixture(name="live_plot_type")
def fixture_plot_type():
    """return lines type"""
    return LivePlotType.LINES


@pytest.fixture(name="live_plot_labels")
def fixture_plot_labels():
    """return valid labels instance"""
    return LivePlotLabels()


@pytest.fixture(name="live_plot_axis")
def fixture_plot_axis():
    """return valid axis instance"""
    return LivePlotAxis()


def test_live_plots_constructor():
    """test live plot creation"""
    live_plots = LivePlots()
    assert isinstance(live_plots, LivePlots)


@patch("websockets.connect", autospec=True)
def test_live_plots_add_plot(
    mocked_websockets_connect: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Tests LivePlots add plot functionality"""

    loop = get_current_event_loop_or_create()

    mocked_connection_future = loop.create_future()
    mocked_connection_future.set_result(MagicMock())
    mocked_websockets_connect.return_value = mocked_connection_future

    live_plots = LivePlots()
    expected_live_plot = LivePlot(
        plot_id=1,
        plot_type=live_plot_type,
        websocket_url="server/demo-url",
        labels=live_plot_labels,
        axis=live_plot_axis,
    )
    asyncio.run(expected_live_plot.start_up())

    asyncio.run(
        live_plots.create_live_plot(
            plot_id=1,
            plot_type=live_plot_type,
            websocket_url="server/demo-url",
            labels=live_plot_labels,
            axis=live_plot_axis,
        )
    )
    created_live_plot = live_plots._get_live_plot(plot_id=1)

    mocked_websockets_connect.assert_called_with("server/demo-url")
    assert expected_live_plot == created_live_plot


@patch("qiboconnection.models.live_plot.LivePlot.send_data", autospec=True)
@patch("qiboconnection.models.live_plot.websockets.connect", autospec=True)
def test_live_plots_send_data(
    mocked_websockets_connect: MagicMock,
    live_plot_send_data: MagicMock,
    live_plot_type: LivePlotType,
    live_plot_labels: LivePlotLabels,
    live_plot_axis: LivePlotAxis,
):
    """Tests the LivePlots send_data functionality, mocking the inferior LivePlot layer."""

    loop = get_current_event_loop_or_create()

    mocked_connection_future = loop.create_future()
    mocked_connection_future.set_result(MagicMock())
    mocked_websockets_connect.return_value = mocked_connection_future

    live_plots = LivePlots()
    asyncio.run(
        live_plots.create_live_plot(
            plot_id=1,
            plot_type=live_plot_type,
            websocket_url="server/demo-url",
            labels=live_plot_labels,
            axis=live_plot_axis,
        )
    )

    asyncio.run(live_plots.send_data(plot_id=1, x=1, y=1, z=None))

    expected_sent_data_packet = LivePlotPacket.build_packet(
        plot_id=1, plot_type=live_plot_type, x=1, y=1, z=None, labels=live_plot_labels, axis=live_plot_axis
    )

    live_plot_send_data.assert_called_with(self=live_plots._get_live_plot(plot_id=1), data=expected_sent_data_packet)


def test_check_data_and_plot_type_compatibility_with_ok_case(
    live_plot_type: LivePlotType, live_plot_labels: LivePlotLabels, live_plot_axis: LivePlotAxis
):
    """test live plot `CheckDataAndPlotTypeCompatibility` works with base case"""
    live_plot = LivePlot(
        plot_id=1,
        plot_type=live_plot_type,
        websocket_url="server/demo-url",
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=live_plot_type,
        x=1,
        y=2,
        z=None,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    test_result = LivePlots.CheckDataAndPlotTypeCompatibility(method=ok_method)(
        live_plot=live_plot, data_packet=data_packet
    )

    assert test_result == "OK"


def test_check_data_and_plot_type_compatibility_rises_attribute_error():
    """test live plot `CheckDataAndPlotTypeCompatibility` works as expected in wrong attr err case"""

    with pytest.raises(AttributeError) as e_info:
        _ = LivePlots.CheckDataAndPlotTypeCompatibility(method=ok_method)(live_plot=None, data_packet=None)

    assert e_info.value.args[0] == "live_plot and point info are required."


def test_check_data_and_plot_type_compatibility_rises_value_error_for_lines(
    live_plot_labels: LivePlotLabels, live_plot_axis: LivePlotAxis
):
    """test live plot `CheckDataAndPlotTypeCompatibility` works as expected in wrong val err case"""
    live_plot = LivePlot(
        plot_id=1,
        plot_type=LivePlotType.LINES,
        websocket_url="server/demo-url",
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=LivePlotType.SCATTER3D,
        x=1,
        y=2,
        z=3,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    with pytest.raises(ValueError) as e_info:
        _ = LivePlots.CheckDataAndPlotTypeCompatibility(method=ok_method)(live_plot=live_plot, data_packet=data_packet)

    assert e_info.value.args[0] == "LINES and SCATTER plots accept exactly x and y values."


def test_check_data_and_plot_type_compatibility_rises_value_error_for_scatter3d(
    live_plot_labels: LivePlotLabels, live_plot_axis: LivePlotAxis
):
    """test live plot `CheckDataAndPlotTypeCompatibility` works as expected in wrong val err case"""
    live_plot = LivePlot(
        plot_id=1,
        plot_type=LivePlotType.SCATTER3D,
        websocket_url="server/demo-url",
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=LivePlotType.LINES,
        x=1,
        y=2,
        z=None,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    with pytest.raises(ValueError) as e_info:
        _ = LivePlots.CheckDataAndPlotTypeCompatibility(method=ok_method)(live_plot=live_plot, data_packet=data_packet)

    assert e_info.value.args[0] == "SCATTER3D plots accept exactly x, y and z values."


def test_check_data_and_plot_type_compatibility_rises_value_error_for_heatmap(
    live_plot_labels: LivePlotLabels, live_plot_axis: LivePlotAxis
):
    """test live plot `CheckDataAndPlotTypeCompatibility` works as expected in wrong val err case"""
    live_plot = LivePlot(
        plot_id=1,
        plot_type=LivePlotType.HEATMAP,
        websocket_url="server/demo-url",
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=LivePlotType.LINES,
        x=1,
        y=2,
        z=None,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    with pytest.raises(ValueError) as e_info:
        _ = LivePlots.CheckDataAndPlotTypeCompatibility(method=ok_method)(live_plot=live_plot, data_packet=data_packet)

    assert e_info.value.args[0] == "HEATMAP plots accept exactly x, y and z values."


def test_check_data_and_plot_type_compatibility_rises_value_error_for_heatmap_axis(
    live_plot_labels: LivePlotLabels, live_plot_axis: LivePlotAxis
):
    """test live plot `CheckDataAndPlotTypeCompatibility` works as expected in wrong val err case"""
    live_plot = LivePlot(
        plot_id=1,
        plot_type=LivePlotType.HEATMAP,
        websocket_url="server/demo-url",
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    data_packet = LivePlotPacket.build_packet(
        plot_id=1,
        plot_type=LivePlotType.SCATTER3D,
        x=1,
        y=2,
        z=3,
        labels=live_plot_labels,
        axis=live_plot_axis,
    )

    with pytest.raises(ValueError) as e_info:
        _ = LivePlots.CheckDataAndPlotTypeCompatibility(method=ok_method)(live_plot=live_plot, data_packet=data_packet)

    assert e_info.value.args[0] == "HEATMAP plots need to have axis provided on live-plot creation."



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" AlgorithmType enum """
from .str_enum import StrEnum


class AlgorithmType(StrEnum):
    """Type of algorithm

    Args:
        enum (str):name types of the initial values.
        Supported values are: 'Gate-Based Circuit' and 'Annealing'
    """

    GATE_BASED = "Gate-Based Circuit"
    ANNEALING = "Annealing"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Live Plot Typing """
import json
from abc import ABC
from dataclasses import asdict, dataclass
from functools import partial
from typing import Callable, List, TypedDict, cast

import numpy as np
import numpy.typing as npt

from .enums import LivePlotType


class UnitPoint(TypedDict):
    """Basic point"""

    x: float
    y: float
    z: float | None
    idx: int | None
    idy: int | None


@dataclass
class LivePlotLabels(ABC):
    """Class for holding different labels for the plots: title, axis names, etc."""

    title: str | None = None
    x_label: str | None = None
    y_label: str | None = None
    z_label: str | None = None


@dataclass
class LivePlotAxis(ABC):
    """Class for holding different axis marks for the plots"""

    x_axis: npt.NDArray | List | None = None
    y_axis: npt.NDArray | List | None = None


class LivePlotPoints(ABC):
    """Information about the points we intend to plot in each message we sent over the live-plotting ws."""

    def __init__(
        self,
        x: int | float | list[int] | list[float] | npt.NDArray[np.int_ | np.float_],
        y: int | float | list[int] | list[float] | npt.NDArray[np.int_ | np.float_],
        z: int | float | list[int] | list[float] | npt.NDArray[np.int_ | np.float_] | None = None,
        idx: int | list[int] | npt.NDArray[np.int_] | None = None,
        idy: int | list[int] | npt.NDArray[np.int_] | None = None,
    ):
        self._x = x
        self._y = y
        self._z = z
        self._idx = idx
        self._idy = idy
        self._points: list[UnitPoint] = []
        self._parse_to_points(x=x, y=y, z=z, idx=idx, idy=idy)

    @property
    def x(self):
        """x getter."""
        return self._x

    @property
    def y(self):
        """y getter."""
        return self._y

    @property
    def z(self):
        """z getter."""
        return self._z

    @property
    def idx(self):
        """idx getter."""
        return self._idx

    @property
    def idy(self):
        """idy getter."""
        return self._idy

    def _parse_to_points(
        self,
        x: int | float | list[int] | list[float] | npt.NDArray[np.int_ | np.float_],
        y: int | float | list[int] | list[float] | npt.NDArray[np.int_ | np.float_],
        z: int | float | list[int] | list[float] | npt.NDArray[np.int_ | np.float_] | None = None,
        idx: int | list[int] | npt.NDArray[np.int_] | None = None,
        idy: int | list[int] | npt.NDArray[np.int_] | None = None,
    ):
        """
        Gets a point or a list of points for x, y, and z and parses them to a list of dicts each containing ONE value
         for each of x, y, z.
        Args:
            x: x value of the point(s) to draw
            y: y value of the point(s) to draw
            z: z value of the point(s) to draw
        Raises:
            ValueError: Arguments provided must be of the same type: floats or lists
        """
        if all(isinstance(arg, int | float | None) for arg in [x, y, z]):
            point = UnitPoint(x=cast(float, x), y=cast(float, y), z=None, idx=None, idy=None)
            if z is not None:
                point["z"] = cast(float, z)
            if idx is not None:
                point["idx"] = cast(int, idx)
            if idy is not None:
                point["idy"] = cast(int, idy)
            self._points.append(point)
        elif all(isinstance(arg, np.ndarray | list | None) for arg in [x, y, z]):
            x, y = cast(list, x), cast(list, y)
            for i, _ in enumerate(x):
                # pylint: disable=unnecessary-list-index-lookup
                point = UnitPoint(x=x[i], y=y[i], z=None, idx=None, idy=None)
                if z is not None:
                    point["z"] = cast(list, z)[i]
                if idx is not None:
                    point["idx"] = cast(list, idx)[i]
                if idy is not None:
                    point["idy"] = cast(list, idy)[i]
                self._points.append(point)
        else:
            raise ValueError("Arguments provided must be of the same type: floats or lists")

    def to_scatter(self):
        """Returns the points in the shape of [{x:x0, y:y0, f:f0}, {x:x1, y:y1, f:f1}, ...]"""
        return self._points

    def __eq__(self, other):
        if isinstance(other, LivePlotPoints):
            return all(
                [
                    self.x == other.x,
                    self.y == other.y,
                    self.z == other.z,
                    self._idx == other._idx,
                    self._idy == other._idy,
                ]
            )
        return False


def _ensure_packet_types(packets: List):
    """Ensures all elements are valid LivePlotPackets"""
    if not all(isinstance(packet, LivePlotPacket) for packet in packets):
        raise ValueError("Not all packets were LivePlotPackets")


def _ensure_packet_compatibility(packets: List):
    """Ensures all elements of packets are from a same compatible Liveplot"""
    same_plot_id = (packet.plot_id == packets[0].plot_id for packet in packets)
    same_plot_type = (packet.plot_type == packets[0].plot_type for packet in packets)
    same_labels = (packet.labels == packets[0].labels for packet in packets)
    same_axis = (packet.axis == packets[0].axis for packet in packets)
    if not (all(same_plot_id) and all(same_plot_type) and all(same_labels) and all(same_axis)):
        raise ValueError("Trying to agglutinate data packets with different information")


@dataclass
class LivePlotPacket(ABC):
    """Packet with the needed information for sending in each message we want to throw over the live-plotting ws."""

    plot_id: int
    plot_type: LivePlotType
    data: LivePlotPoints
    labels: LivePlotLabels
    axis: LivePlotAxis

    class ParseDataIfNeeded:
        """Function decorator used to add extra kwargs to the provided points before sending them, and to downcast
        provided data into json-serializable datatypes."""

        def __init__(self, method: Callable):
            self._method = method

        def __get__(self, obj, objtype):
            """Support instance methods."""
            return partial(self.__call__, obj)

        def __call__(self, *args, **kwargs):
            """
            Args:
                method (Callable): Class method.
            """

            downcasted_kwargs = self._downcast_ndarrays_to_lists(**kwargs)
            extended_kwargs = self._add_index_args_for_heatmaps(**downcasted_kwargs)

            return self._method(*args, **extended_kwargs)

        def _downcast_ndarrays_to_lists(self, **kwargs):
            """Downcasts all provided numpy arrays to lists"""
            for key, value in kwargs.items():
                if isinstance(value, np.ndarray):
                    kwargs[key] = value.tolist()

            return kwargs

        def _add_index_args_for_heatmaps(self, **kwargs):
            plot_type: LivePlotAxis = kwargs.get("plot_type")
            if plot_type == LivePlotType.HEATMAP:
                axis: LivePlotAxis = kwargs.get("axis")

                x: List[float | int] | float | int = kwargs.get("x")
                y: List[float | int] | float | int = kwargs.get("y")
                idx = None
                idy = None

                if isinstance(x, list) and isinstance(y, list):
                    idx = [int(np.where(np.unique(axis.x_axis) == i)[0][0]) for i in x]
                    idy = [int(np.where(np.unique(axis.y_axis) == i)[0][0]) for i in y]
                if isinstance(x, (float, int)) and isinstance(y, (float, int)):
                    idx = int(np.where(np.unique(axis.x_axis) == x)[0][0])
                    idy = int(np.where(np.unique(axis.y_axis) == y)[0][0])

                return {**kwargs, "idx": idx, "idy": idy}
            return kwargs

    @classmethod
    @ParseDataIfNeeded
    def build_packet(
        cls,
        plot_id: int,
        plot_type: LivePlotType,
        labels: LivePlotLabels,
        axis: LivePlotAxis,
        x: npt.NDArray[np.float_ | np.int_] | list[float] | list[int] | float | int,
        y: npt.NDArray[np.float_ | np.int_] | list[float] | list[int] | float | int,
        z: npt.NDArray[np.float_ | np.int_] | list[float] | list[int] | float | int | None,
        idx: npt.NDArray[np.int_] | list[int] | int | None = None,
        idy: npt.NDArray[np.int_] | list[int] | int | None = None,
    ):
        """Convenience constructor"""
        return cls(
            plot_id=plot_id,
            plot_type=plot_type,
            labels=labels,
            axis=axis,
            data=LivePlotPoints(x=x, y=y, z=z, idx=idx, idy=idy),
        )

    def to_dict(self) -> dict:
        """
        Serializes the information of the class to a dict ready to be sent via ws.
        """
        return {
            "plot_id": self.plot_id,
            "plot_type": self.plot_type.value,
            "labels": asdict(self.labels),
            "data": self.data.to_scatter(),
        }

    def to_json(self) -> str:
        """
        Serializes the info of the class in a json-formatted string
        """
        return json.dumps(self.to_dict())

    @classmethod
    def agglutinate(cls, packets: List):
        """Build a single LivePlotPacket from a list of LivePlotPackets with the data merged"""

        if not packets:
            return None

        _ensure_packet_types(packets=packets)
        _ensure_packet_compatibility(packets=packets)

        extended_x: list[int] | list[float] = list[float]()
        extended_y: list[int] | list[float] = list[float]()
        extended_z: list[int] | list[float] = list[float]()
        extended_idx: list[int] = list[int]()
        extended_idy: list[int] = list[int]()
        extended_x.extend((packet.data.x for packet in packets))
        extended_y.extend((packet.data.y for packet in packets))
        extended_z.extend((packet.data.z for packet in packets))
        extended_idx.extend((packet.data.idx for packet in packets))
        extended_idy.extend((packet.data.idy for packet in packets))
        return cls(
            plot_id=packets[0].plot_id,
            plot_type=packets[0].plot_type,
            data=LivePlotPoints(x=extended_x, y=extended_y, z=extended_z, idx=extended_idx, idy=extended_idy),
            labels=packets[0].labels,
            axis=packets[0].axis,
        )



"""API testing"""
import ast
import asyncio
import base64
import json
from dataclasses import asdict
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
import responses  # pylint: disable=import-error
from qibo import gates
from qibo.models import Circuit

from qiboconnection.api import API
from qiboconnection.connection import ConnectionConfiguration
from qiboconnection.errors import ConnectionException, RemoteExecutionException
from qiboconnection.models.devices.devices import Devices
from qiboconnection.models.devices.util import create_device
from qiboconnection.models.job_listing import JobListing
from qiboconnection.models.runcard import Runcard
from qiboconnection.models.saved_experiment import SavedExperiment
from qiboconnection.models.saved_experiment_listing import SavedExperimentListing
from qiboconnection.typings.job_data import JobData
from qiboconnection.typings.responses import PlottingResponse
from tests.data.web_responses.job import JobResponse

from .data import experiment_dict, results_dict, runcard_dict, web_responses
from .utils import get_current_event_loop_or_create


def test_api_constructor(mocked_api: API):
    """Test API class constructor"""
    assert isinstance(mocked_api, API)


@patch("qiboconnection.api.API.__init__", autospec=True)
def test_api_login(mocked_api_init: MagicMock):
    """Tests user utility constructor Login calls __init__ with the correct information,"""
    mocked_api_init.return_value = None
    _USERNAME = "test-name"
    _API_KEY = "test-key"

    _ = API.login(username=_USERNAME, api_key=_API_KEY)

    provided_user_info = mocked_api_init.call_args_list[0][1]["configuration"]
    assert isinstance(provided_user_info, ConnectionConfiguration)
    assert provided_user_info.username == _USERNAME
    assert provided_user_info.api_key == _API_KEY


def test_jobs(mocked_api: API):
    """Test jobs property"""
    assert mocked_api.jobs == []


def test_last_job(mocked_api: API):
    """Test last_job property"""
    with pytest.raises(IndexError):
        _ = mocked_api.last_job


@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_last_saved_experiment(mocked_web_call: MagicMock, mocked_api: API):
    """Test last_saved_experiment property"""
    mocked_web_call.return_value = web_responses.saved_experiments.create_response

    assert mocked_api._last_saved_experiment is None
    name = "MyDemoExperiment"
    description = "A test saved experiment"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"
    favourite = True
    _ = mocked_api._save_experiment(
        name=name,
        description=description,
        experiment_dict=experiment_dict,
        results_dict=results_dict,
        device_id=device_id,
        user_id=user_id,
        qililab_version=qililab_version,
        favourite=favourite,
    )
    assert isinstance(mocked_api._last_saved_experiment, SavedExperiment)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_last_saved_experiment_listing(mocked_web_call: MagicMock, mocked_api: API):
    """Test last_saved_experiment_listing property"""
    mocked_web_call.return_value = web_responses.saved_experiments.retrieve_listing_response

    assert mocked_api._last_saved_experiment_listing is None
    saved_experiments_list = mocked_api._list_saved_experiments()
    assert isinstance(mocked_api._last_saved_experiment_listing, SavedExperimentListing)
    assert mocked_api._last_saved_experiment_listing == saved_experiments_list


@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_last_runcard(mocked_web_call: MagicMock, mocked_api: API):
    """Test last_runcard property"""
    mocked_web_call.return_value = web_responses.runcards.create_response

    assert mocked_api.last_runcard is None
    name = "MyDemoRuncard"
    description = "A test runcard"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"
    _ = mocked_api.save_runcard(
        name=name,
        description=description,
        runcard_dict=runcard_dict,
        device_id=device_id,
        user_id=user_id,
        qililab_version=qililab_version,
    )
    assert isinstance(mocked_api.last_runcard, Runcard)


def test_user_id(mocked_api: API):
    """Test last_runcard property"""
    assert mocked_api.user_id == 666


@patch("qiboconnection.connection.Connection.send_get_remote_call", autospec=True)
def test_ping(mocked_web_call: MagicMock, mocked_api: API):
    """Test ping function"""
    mocked_web_call.return_value = web_responses.ping.success_response

    mocked_api.ping()

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=mocked_api._PING_CALL_PATH)


@patch("qiboconnection.connection.Connection.send_get_remote_call", autospec=True)
def test_ping_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Test ping function with failing"""
    mocked_web_call.return_value = web_responses.ping.ise_response

    with pytest.raises(ConnectionException):
        mocked_api.ping()

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=mocked_api._PING_CALL_PATH)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_list_devices(mocked_web_call: MagicMock, mocked_api: API):
    """Test list devices function"""
    mocked_web_call.return_value = web_responses.devices.retrieve_many_response

    device_listing = mocked_api.list_devices()

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=mocked_api._DEVICES_CALL_PATH)
    assert isinstance(device_listing, Devices)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_list_devices_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Test list devices function"""
    mocked_web_call.return_value = web_responses.devices.ise_many_response

    with pytest.raises(RemoteExecutionException):
        _ = mocked_api.list_devices()

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=mocked_api._DEVICES_CALL_PATH)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_select_device_id(mocked_web_call: MagicMock, mocked_api: API):
    """Test list devices function"""
    mocked_web_call.return_value = web_responses.devices.retrieve_response

    mocked_api.select_device_id(device_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._DEVICES_CALL_PATH}/1")
    assert mocked_api._selected_devices is not None
    assert len(mocked_api._selected_devices) == 1
    assert mocked_api._selected_devices[0] == create_device(web_responses.devices.retrieve_response[0])


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_select_device_id_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Test list devices function"""
    mocked_web_call.return_value = web_responses.devices.ise_response

    with pytest.raises(RemoteExecutionException):
        mocked_api.select_device_id(device_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._DEVICES_CALL_PATH}/1")
    assert mocked_api._selected_devices == []


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_select_device_ids(mocked_web_call: MagicMock, mocked_api: API):
    """Test list devices function"""
    mocked_web_call.return_value = web_responses.devices.retrieve_response

    mocked_api.select_device_ids(device_ids=[1])

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._DEVICES_CALL_PATH}/1")
    assert mocked_api._selected_devices is not None
    assert len(mocked_api._selected_devices) == 1
    assert mocked_api._selected_devices[0] == create_device(web_responses.devices.retrieve_response[0])


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_select_device_ids_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Test list devices function"""
    mocked_web_call.return_value = web_responses.devices.ise_response

    with pytest.raises(RemoteExecutionException):
        mocked_api.select_device_ids(device_ids=[1])

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._DEVICES_CALL_PATH}/1")
    assert mocked_api._selected_devices == []


@patch("websockets.connect", autospec=True)
@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_create_live_plot(mocked_web_call: MagicMock, mocked_websockets_connect: MagicMock, mocked_api: API):
    """Test the creation of a liveplot using the api"""

    loop = get_current_event_loop_or_create()
    mocked_connection_future = loop.create_future()
    mocked_connection_future.set_result(MagicMock())
    mocked_websockets_connect.return_value = mocked_connection_future

    mocked_web_call.return_value = PlottingResponse(websocket_url="test_url", plot_id=1).to_dict(), 200

    plot_id = asyncio.run(mocked_api._create_liveplot())

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=mocked_api._LIVE_PLOTTING_PATH, data={})
    mocked_websockets_connect.assert_called_with("test_url")
    assert plot_id == 1


@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_create_live_plot_with_unexpected_response(mocked_web_call: MagicMock, mocked_api: API):
    """Test the creation of a liveplot using the api"""

    mocked_web_call.return_value = {}, 400

    with pytest.raises(RemoteExecutionException):
        _ = asyncio.run(mocked_api._create_liveplot())


@patch("qiboconnection.models.live_plot.LivePlot.send_data", autospec=True)
@patch("websockets.connect", autospec=True)
@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_send_plot_points(
    mocked_web_call: MagicMock,
    mocked_websockets_connect: MagicMock,
    mocked_live_plot_send_data: MagicMock,
    mocked_api: API,
):
    """Test the creation of a liveplot using the api"""

    loop = get_current_event_loop_or_create()
    mocked_connection_future = loop.create_future()
    mocked_connection_future.set_result(MagicMock())
    mocked_websockets_connect.return_value = mocked_connection_future

    mocked_web_call.return_value = PlottingResponse(websocket_url="test_url", plot_id=1).to_dict(), 200

    plot_id = asyncio.run(mocked_api._create_liveplot())

    asyncio.run(mocked_api._send_plot_points(plot_id=plot_id, x=0, y=0))

    mocked_live_plot_send_data.assert_called()


@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_save_experiment(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.save_experiment() method"""
    mocked_web_call.return_value = web_responses.saved_experiments.create_response

    name = "MyDemoExperiment"
    description = "A test saved experiment"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"
    favourite = True

    saved_experiment_id = mocked_api._save_experiment(
        name=name,
        description=description,
        experiment_dict=experiment_dict,
        results_dict=results_dict,
        device_id=device_id,
        user_id=user_id,
        qililab_version=qililab_version,
        favourite=favourite,
    )

    mocked_web_call.assert_called()
    assert saved_experiment_id == web_responses.saved_experiments.create_response[0]["saved_experiment_id"]


@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_save_experiment_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.save_experiment() method"""
    mocked_web_call.return_value = web_responses.saved_experiments.ise_response

    name = "MyDemoExperiment"
    description = "A test saved experiment"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"
    favourite = True
    with pytest.raises(RemoteExecutionException):
        _ = mocked_api._save_experiment(
            name=name,
            description=description,
            experiment_dict=experiment_dict,
            results_dict=results_dict,
            device_id=device_id,
            user_id=user_id,
            qililab_version=qililab_version,
            favourite=favourite,
        )
    mocked_web_call.assert_called()


@pytest.mark.parametrize("favourites", [False, True])
@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_list_jobs(mocked_web_call: MagicMock, favourites: bool, mocked_api: API):
    """Tests API.list_jobs() method"""
    mocked_web_call.return_value = web_responses.job_response.retrieve_job_listing_response

    jobs_list = mocked_api.list_jobs(favourites=favourites)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection, path=mocked_api._JOBS_CALL_PATH, params={"favourites": favourites}
    )
    assert isinstance(jobs_list, JobListing)
    assert isinstance(jobs_list.dataframe, pd.DataFrame)


@pytest.mark.parametrize("favourites", [False, True])
@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_list_saved_experiments(mocked_web_call: MagicMock, favourites: bool, mocked_api: API):
    """Tests API.list_saved_experiments() method"""
    mocked_web_call.return_value = web_responses.saved_experiments.retrieve_listing_response

    saved_experiments_list = mocked_api._list_saved_experiments(favourites=favourites)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection, path=mocked_api._SAVED_EXPERIMENTS_CALL_PATH, params={"favourites": favourites}
    )
    assert isinstance(saved_experiments_list, SavedExperimentListing)
    assert isinstance(saved_experiments_list.dataframe, pd.DataFrame)


@pytest.mark.parametrize("favourites", [False, True])
@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_list_saved_experiments_ise(mocked_web_call: MagicMock, favourites: bool, mocked_api: API):
    """Tests API.list_saved_experiments() method"""
    mocked_web_call.return_value = web_responses.saved_experiments.ise_listing_response

    with pytest.raises(RemoteExecutionException):
        _ = mocked_api._list_saved_experiments(favourites=favourites)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection, path=mocked_api._SAVED_EXPERIMENTS_CALL_PATH, params={"favourites": favourites}
    )


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_saved_experiment(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_saved_experiment() method"""
    mocked_web_call.return_value = web_responses.saved_experiments.retrieve_response

    saved_experiment = mocked_api._get_saved_experiment(saved_experiment_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1")
    assert isinstance(saved_experiment, SavedExperiment)


@pytest.mark.parametrize(
    "web_job_response",
    [JobResponse.retrieve_job_response_1, JobResponse.retrieve_job_response_2, JobResponse.retrieve_job_response_3],
)
@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_job(mocked_web_call: MagicMock, mocked_api: API, web_job_response: dict):
    """Tests API.get_job() method."""
    mocked_web_call.return_value = web_job_response

    job_data = mocked_api.get_job(job_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._JOBS_CALL_PATH}/1")
    assert isinstance(job_data, JobData)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_job_exception(mocked_api_call: MagicMock, mocked_api: API):
    """Tests API.get_result() method with non-existent job id"""

    # Define the behavior of the mocked function to raise the RemoteExecutionException
    mocked_api_call.side_effect = RemoteExecutionException("The job does not exist!", status_code=400)

    with pytest.raises(RemoteExecutionException, match="The job does not exist!"):
        # Call the function that should raise the exception
        mocked_api.get_result(job_id=0)

    # Assert that the mocked function was called with correct arguments
    mocked_api_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._JOBS_CALL_PATH}/{0}")


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_saved_experiment_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_saved_experiment() method"""
    mocked_web_call.return_value = web_responses.saved_experiments.ise_response

    with pytest.raises(RemoteExecutionException):
        _ = mocked_api._get_saved_experiment(saved_experiment_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1")


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_saved_experiments(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_saved_experiments() method"""
    mocked_web_call.return_value = web_responses.saved_experiments.retrieve_response

    saved_experiment = mocked_api._get_saved_experiments(saved_experiment_ids=[1])

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1")
    assert isinstance(saved_experiment[0], SavedExperiment)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_saved_experiments_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_saved_experiments() method"""
    mocked_web_call.return_value = web_responses.saved_experiments.ise_response

    with pytest.raises(RemoteExecutionException):
        _ = mocked_api._get_saved_experiments(saved_experiment_ids=[1])

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1")


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_fav_saved_experiment(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.fav_saved_experiment() with ONE experiment"""
    mocked_web_call.return_value = web_responses.saved_experiments.update_response

    mocked_api._fav_saved_experiment(saved_experiment_id=1)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1",
        data={"favourite": True, "user_id": mocked_api.user_id},
    )


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_fav_saved_experiment_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.fav_saved_experiment() with ONE experiment"""
    mocked_web_call.return_value = web_responses.saved_experiments.ise_response

    with pytest.raises(RemoteExecutionException):
        mocked_api._fav_saved_experiment(saved_experiment_id=1)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1",
        data={"favourite": True, "user_id": mocked_api.user_id},
    )


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_fav_saved_experiments(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.fav_saved_experiments() method with a LIST of experiments"""
    mocked_web_call.return_value = web_responses.saved_experiments.update_response

    mocked_api._fav_saved_experiments(saved_experiment_ids=[1])

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1",
        data={"favourite": True, "user_id": mocked_api.user_id},
    )


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_fav_saved_experiments_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.fav_saved_experiments() method with a LIST of experiments"""
    mocked_web_call.return_value = web_responses.saved_experiments.ise_response

    with pytest.raises(RemoteExecutionException):
        mocked_api._fav_saved_experiments(saved_experiment_ids=[1])

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1",
        data={"favourite": True, "user_id": mocked_api.user_id},
    )


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_unfav_saved_experiment(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.unfav_saved_experiment() method with ONE experiment"""
    mocked_web_call.return_value = web_responses.saved_experiments.update_response

    mocked_api._unfav_saved_experiment(saved_experiment_id=1)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1",
        data={"favourite": False, "user_id": mocked_api.user_id},
    )


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_unfav_saved_experiment_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.unfav_saved_experiment() method with ONE experiment"""
    mocked_web_call.return_value = web_responses.saved_experiments.ise_response

    with pytest.raises(RemoteExecutionException):
        mocked_api._unfav_saved_experiment(saved_experiment_id=1)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1",
        data={"favourite": False, "user_id": mocked_api.user_id},
    )


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_unfav_saved_experiments(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.unfav_saved_experiments() method with a LIST of experiments"""
    mocked_web_call.return_value = web_responses.saved_experiments.update_response

    mocked_api._unfav_saved_experiments(saved_experiment_ids=[1])

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1",
        data={"favourite": False, "user_id": mocked_api.user_id},
    )


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_unfav_saved_experiments_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.unfav_saved_experiments() method with a LIST of experiments"""
    mocked_web_call.return_value = web_responses.saved_experiments.ise_response

    with pytest.raises(RemoteExecutionException):
        mocked_api._unfav_saved_experiments(saved_experiment_ids=[1])

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._SAVED_EXPERIMENTS_CALL_PATH}/1",
        data={"favourite": False, "user_id": mocked_api.user_id},
    )


@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_save_runcard(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.save_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.create_response

    name = "MyDemoRuncard"
    description = "A test runcard"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"

    runcard_id = mocked_api.save_runcard(
        name=name,
        description=description,
        runcard_dict=runcard_dict,
        device_id=device_id,
        user_id=user_id,
        qililab_version=qililab_version,
    )

    mocked_web_call.assert_called()
    assert runcard_id == web_responses.runcards.create_response[0]["runcard_id"]


@patch("qiboconnection.connection.Connection.send_post_auth_remote_api_call", autospec=True)
def test_save_runcard_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.save_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.ise_response

    name = "MyDemoRuncard"
    description = "A test runcard"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"

    with pytest.raises(RemoteExecutionException):
        _ = mocked_api.save_runcard(
            name=name,
            description=description,
            runcard_dict=runcard_dict,
            device_id=device_id,
            user_id=user_id,
            qililab_version=qililab_version,
        )

    mocked_web_call.assert_called()


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_runcard(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_runcard() method using runcard id"""
    mocked_web_call.return_value = web_responses.runcards.retrieve_response

    runcard = mocked_api.get_runcard(runcard_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._RUNCARDS_CALL_PATH}/1")
    assert isinstance(runcard, Runcard)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_runcard_by_name(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_runcard() method using runcard name"""
    mocked_web_call.return_value = web_responses.runcards.retrieve_response

    runcard = mocked_api.get_runcard(runcard_name="DEMO_RUNCARD")

    mocked_web_call.assert_called_with(
        self=mocked_api._connection, path=f"{mocked_api._RUNCARDS_CALL_PATH}/by_keys", params={"name": "DEMO_RUNCARD"}
    )
    assert isinstance(runcard, Runcard)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_runcard_with_redundant_info(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_runcard() method fails when providing name and id at the same time"""
    mocked_web_call.return_value = web_responses.runcards.retrieve_response

    with pytest.raises(ValueError):
        _ = mocked_api.get_runcard(runcard_id=1, runcard_name="TEST_RUNCARD")

    mocked_web_call.assert_not_called()


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_runcard_with_insufficient_info(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_runcard() method fails if not name nor id are provided"""
    mocked_web_call.return_value = web_responses.runcards.retrieve_response

    with pytest.raises(ValueError):
        _ = mocked_api.get_runcard()

    mocked_web_call.assert_not_called()


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_runcard_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.get_runcard() method when server response raises an error"""
    mocked_web_call.return_value = web_responses.runcards.ise_response

    with pytest.raises(RemoteExecutionException):
        _ = mocked_api.get_runcard(runcard_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._RUNCARDS_CALL_PATH}/1")


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_list_runcards(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.list_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.retrieve_many_response

    runcards = mocked_api.list_runcards()

    mocked_web_call.assert_called()
    assert isinstance(runcards[0], Runcard)


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_list_runcards_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.list_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.ise_many_response

    with pytest.raises(RemoteExecutionException):
        _ = mocked_api.list_runcards()

    mocked_web_call.assert_called()


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_update_runcard(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.update_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.update_response

    runcard_id = 1
    name = "MyDemoRuncard"
    description = "A test runcard"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"
    modified_runcard = Runcard(
        id=runcard_id,
        name=name,
        description=description,
        runcard=runcard_dict,
        device_id=device_id,
        user_id=user_id,
        qililab_version=qililab_version,
    )

    updated_runcard = mocked_api.update_runcard(runcard=modified_runcard)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._RUNCARDS_CALL_PATH}/1",
        data=asdict(modified_runcard.runcard_request()),
    )
    assert isinstance(updated_runcard, Runcard)


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_update_runcard_with_no_id(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.update_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.update_response

    name = "MyDemoRuncard"
    description = "A test runcard"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"
    modified_runcard = Runcard(
        id=None,
        name=name,
        description=description,
        runcard=runcard_dict,
        device_id=device_id,
        user_id=user_id,
        qililab_version=qililab_version,
    )

    with pytest.raises(ValueError):
        _ = mocked_api.update_runcard(runcard=modified_runcard)

    mocked_web_call.assert_not_called()


@patch("qiboconnection.connection.Connection.send_put_auth_remote_api_call", autospec=True)
def test_update_runcard_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.update_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.ise_response

    runcard_id = 1
    name = "MyDemoRuncard"
    description = "A test runcard"
    device_id = 1
    user_id = 1
    qililab_version = "0.0.0"
    modified_runcard = Runcard(
        id=runcard_id,
        name=name,
        description=description,
        runcard=runcard_dict,
        device_id=device_id,
        user_id=user_id,
        qililab_version=qililab_version,
    )

    with pytest.raises(RemoteExecutionException):
        _ = mocked_api.update_runcard(runcard=modified_runcard)

    mocked_web_call.assert_called_with(
        self=mocked_api._connection,
        path=f"{mocked_api._RUNCARDS_CALL_PATH}/1",
        data=asdict(modified_runcard.runcard_request()),
    )


@patch("qiboconnection.connection.Connection.send_delete_auth_remote_api_call", autospec=True)
def test_delete_runcard(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.delete_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.delete_response

    mocked_api.delete_runcard(runcard_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._RUNCARDS_CALL_PATH}/1")


@patch("qiboconnection.connection.Connection.send_delete_auth_remote_api_call", autospec=True)
def test_delete_runcard_ise(mocked_web_call: MagicMock, mocked_api: API):
    """Tests API.delete_runcard() method"""
    mocked_web_call.return_value = web_responses.runcards.ise_response

    with pytest.raises(RemoteExecutionException):
        mocked_api.delete_runcard(runcard_id=1)

    mocked_web_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._RUNCARDS_CALL_PATH}/1")


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call", autospec=True)
def test_get_result_exception(mocked_api_call: MagicMock, mocked_api: API):
    """Tests API.get_result() method with non-existent job id."""

    # Define the behavior of the mocked function to raise the RemoteExecutionException
    mocked_api_call.side_effect = RemoteExecutionException("The job does not exist!", status_code=400)

    with pytest.raises(RemoteExecutionException, match="The job does not exist!"):
        # Call the function that should raise the exception
        mocked_api.get_result(job_id=0)

    # Assert that the mocked function was called with correct arguments
    mocked_api_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._JOBS_CALL_PATH}/{0}")


@patch("qiboconnection.connection.Connection.send_get_auth_remote_api_call_all_pages", autospec=True)
def test_no_devices_selected_exception(mocked_api_call: MagicMock, mocked_api: API):
    """Tests API.execute() method with no devices selected"""

    # Define the behavior of the mocked function to raise the ValueError
    mocked_api_call.side_effect = ValueError("No devices were selected for execution.")

    with pytest.raises(ValueError, match="No devices were selected for execution."):
        # Here, a circuit should be passed which will be ignored as the call is mocked
        mocked_api.execute(device_ids=None)

    # Check that the mocked function was not called
    mocked_api_call.assert_not_called()


@patch("qiboconnection.connection.Connection.send_delete_auth_remote_api_call", autospec=True)
def test_delete_job(mocked_api_call: MagicMock, mocked_api: API):
    """Tests API.delete_job() method"""
    mocked_api_call.return_value = web_responses.job_response.delete_job_response
    mocked_api.delete_job(job_id=0)

    mocked_api_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._JOBS_CALL_PATH}/{0}")


@patch("qiboconnection.connection.Connection.send_delete_auth_remote_api_call", autospec=True)
def test_delete_job_exception(mocked_api_call: MagicMock, mocked_api: API):
    """Tests API.delete_job() method with non-existent job id"""
    # Define the behavior of the mocked function to raise the RemoteExecutionException
    mocked_api_call.return_value = web_responses.job_response.delete_job_response_ise
    with pytest.raises(RemoteExecutionException, match="Job could not be removed."):
        # Call the function that should raise the exception
        mocked_api.delete_job(job_id=0)
    # Assert that the mocked function was called with correct arguments
    mocked_api_call.assert_called_with(self=mocked_api._connection, path=f"{mocked_api._JOBS_CALL_PATH}/{0}")


class TestExecute:
    """Unit tests for the `API.execute` method."""

    r_mock: responses.RequestsMock

    circuit = Circuit(5)
    circuit.add(gates.X(0))
    circuit.add(gates.H(4))
    circuit.add(gates.M(0, 1, 2, 3, 4))

    def setup_method(self):
        """Method executed before each test contained in this class.

        This method mocks the `requests` calls used inside the API.execute method.
        """
        self.r_mock = responses.RequestsMock(assert_all_requests_are_fired=True)
        self.r_mock.start()
        self.r_mock.add(
            method="GET",
            url="https://qilimanjaroqaas.ddns.net:8080/api/v1/devices/9",
            status=200,
            json={
                "status": "offline",
                "device_id": 9,
                "device_name": "Dummy Device",
                "availability": "available",
                "channel_id": 0,
            },
        )
        self.r_mock.add(
            method="POST", url="https://qilimanjaroqaas.ddns.net:8080/api/v1/circuits", status=201, json={"job_id": 0}
        )

    def teardown_method(self):
        """Method executed at the end of each test contained in this class.

        Stops and resets RequestsMock instance. If ``assert_all_requests_are_fired`` is set to ``True``, will raise an
        error if some requests were not processed.
        """
        self.r_mock.stop()
        self.r_mock.reset()

    def test_execute_with_one_circuit(self, mocked_api: API):
        """Test the API.execute method for a single circuit."""
        job_ids = mocked_api.execute(circuit=self.circuit, nshots=1000, device_ids=[9])

        assert job_ids == [0]
        assert len(self.r_mock.calls) == 2
        body = json.loads(self.r_mock.calls[1].request.body.decode())
        assert body["device_id"] == 9
        assert body["number_shots"] == 1000
        assert body["job_type"] == "circuit"
        description = ast.literal_eval(body["description"])
        assert len(description) == 1
        assert (
            base64.urlsafe_b64decode(description[0]).decode() == self.circuit.to_qasm()
        )  # make sure we posted the correct circuit

    def test_execute_with_multiple_circuits(self, mocked_api: API):
        """Test the API.execute method for multiple circuits."""
        job_ids = mocked_api.execute(circuit=[self.circuit] * 10, nshots=1000, device_ids=[9])

        assert job_ids == [0]
        assert len(self.r_mock.calls) == 2
        body = json.loads(self.r_mock.calls[1].request.body.decode())
        assert body["device_id"] == 9
        assert body["number_shots"] == 1000
        assert body["job_type"] == "circuit"
        description = ast.literal_eval(body["description"])
        assert len(description) == 10
        assert all(
            base64.urlsafe_b64decode(d).decode() == self.circuit.to_qasm() for d in description
        )  # make sure we posted the correct circuits



""" Test methods for Runcard typing classes"""

from datetime import datetime, timezone

from qiboconnection.typings.responses.runcard_response import RuncardRequest, RuncardResponse


def test_runcard_request_creation():
    """Tests RuncardRequest creation"""

    runcard_request = RuncardRequest(
        name="runcard",
        description="description",
        user_id=0,
        device_id=0,
        runcard="eyJhIjogMH0=",
        qililab_version="0.0.0",
    )

    assert isinstance(runcard_request, RuncardRequest)


def test_runcard_response_creation():
    """Tests RuncardRequest creation"""

    runcard_response = RuncardResponse(
        name="runcard",
        description="description",
        user_id=0,
        device_id=0,
        runcard="eyJhIjogMH0=",
        qililab_version="0.0.0",
        runcard_id=0,
        created_at=datetime.now(timezone.utc),
        updated_at=datetime.now(timezone.utc),
    )

    assert isinstance(runcard_response, RuncardResponse)



""" Test Environment setting """

import enum
import os

import pytest

from qiboconnection.config import Environment, EnvironmentType, get_environment


def test_environment_constructor():
    """Test Environment class constructor."""
    environment_type = EnvironmentType.DEVELOPMENT

    environment = Environment(environment_type=environment_type)
    assert isinstance(environment, Environment)


def test_environment_constructor_raises_value_error_for_unexpected_environment_types():
    """Test Environment class constructor."""

    class NewEnvironmentType(str, enum.Enum):
        UNEXPECTED = "unexpected"

    with pytest.raises(ValueError) as e_info:
        _ = Environment(environment_type=NewEnvironmentType.UNEXPECTED)  # type: ignore[arg-type]
    assert e_info.value.args[0] == "Environment Type MUST be 'local', 'staging' or 'development'"


def test_staging_environmemt():
    """Test get_environment() returns a STAGING environment when the QIBOCONNECTION_ENVIRONMENT env variable is set to
    'staging'"""
    os.environ["QIBOCONNECTION_ENVIRONMENT"] = EnvironmentType.STAGING.value
    environment = get_environment()
    assert environment.environment_type == EnvironmentType.STAGING


def test_development_environmemt():
    """Test get_environment() returns a DEVELOPMENT environment when the QIBOCONNECTION_ENVIRONMENT env variable is set
    to 'development'"""
    os.environ["QIBOCONNECTION_ENVIRONMENT"] = EnvironmentType.DEVELOPMENT.value
    environment = get_environment()
    assert environment.environment_type == EnvironmentType.DEVELOPMENT


def test_local_environmemt():
    """Test get_environment() returns a LOCAL environment when the QIBOCONNECTION_ENVIRONMENT env variable is set to
    'local'"""
    os.environ["QIBOCONNECTION_ENVIRONMENT"] = EnvironmentType.LOCAL.value
    environment = get_environment()
    assert environment.environment_type == EnvironmentType.LOCAL



""" Test methods for SavedExperimentListing classes"""
import datetime

import pandas as pd

from qiboconnection.models.saved_experiment_listing import SavedExperimentListing, SavedExperimentListingItem
from qiboconnection.typings.responses import SavedExperimentListingItemResponse


def test_saved_experiment_listing_item_creation():
    """Tests SavedExperimentListingItem creation"""

    saved_experiment_listing_item = SavedExperimentListingItem(
        name="name",
        experiment={"a": 0},
        device_id=1,
        user_id=1,
        description="description",
        qililab_version="0.0.0",
        id=0,
        created_at=datetime.datetime.now(),
    )

    assert isinstance(saved_experiment_listing_item, SavedExperimentListingItem)


def test_saved_experiment_listing_item_creation_from_response():
    """Tests SavedExperimentListingItem's from_response() constructor"""

    saved_experiment_listing_item_response = SavedExperimentListingItemResponse(
        name="name",
        experiment="eyJhIjogMH0=",
        device_id=1,
        user_id=1,
        description="description",
        qililab_version="0.0.0",
        id=0,
        created_at=datetime.datetime.now(),
    )

    saved_experiment_listing_item = SavedExperimentListingItem.from_response(
        response=saved_experiment_listing_item_response
    )

    assert isinstance(saved_experiment_listing_item, SavedExperimentListingItem)
    assert saved_experiment_listing_item.experiment == {"a": 0}, "Decoded experiment does not coincide with expected"


def test_saved_experiment_listing_creation():
    """Tests SavedExperimentListing creation"""

    saved_experiment_listing_item = SavedExperimentListingItem(
        name="name",
        experiment={"a": 0},
        device_id=1,
        user_id=1,
        description="description",
        qililab_version="0.0.0",
        id=0,
        created_at=datetime.datetime.now(),
    )

    saved_experiment_listing = SavedExperimentListing(items=[saved_experiment_listing_item])

    assert isinstance(saved_experiment_listing, SavedExperimentListing)
    assert isinstance(saved_experiment_listing.dataframe, pd.DataFrame)


def test_saved_experiment_listing_creation_from_response():
    """Tests SavedExperimentListing creation"""

    saved_experiment_listing_item_response = SavedExperimentListingItemResponse(
        name="name",
        experiment="eyJhIjogMH0=",
        device_id=1,
        user_id=1,
        description="description",
        qililab_version="0.0.0",
        id=0,
        created_at=datetime.datetime.now(),
    )

    saved_experiment_listing = SavedExperimentListing.from_response(
        response_list=[saved_experiment_listing_item_response]
    )

    assert isinstance(saved_experiment_listing, SavedExperimentListing)
    assert isinstance(saved_experiment_listing.dataframe, pd.DataFrame)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" LivePlots class """
from abc import ABC
from dataclasses import dataclass, field
from functools import partial
from typing import Callable

import numpy as np
import numpy.typing as npt

from qiboconnection.typings.live_plot import LivePlotAxis, LivePlotLabels, LivePlotPacket, LivePlotType

from .live_plot import LivePlot


@dataclass
class LivePlots(ABC):
    """Job class to manage the job experiment to be remotely sent"""

    _live_plots: dict = field(default_factory=dict)

    async def create_live_plot(
        self, plot_id: int, plot_type: LivePlotType, websocket_url: str, labels: LivePlotLabels, axis: LivePlotAxis
    ):
        """Creates a new LivePlot with the provided information and appends it to the internal _live_plots dict"""
        self._live_plots[plot_id] = LivePlot(
            plot_id=plot_id,
            plot_type=plot_type,
            websocket_url=websocket_url,
            labels=labels,
            axis=axis,
        )
        await self._live_plots[plot_id].start_up()

    def _get_live_plot(self, plot_id: int) -> LivePlot:
        """
        Internal getter that returns the liveplot instance associatied to the provided plot_id
        Args:
            plot_id: id of the plot to be retrieved.
        """
        return self._live_plots[plot_id]

    class CheckDataAndPlotTypeCompatibility:
        """Function decorator used to check if the points desired to add to a live-plot are compatible with that."""

        def __init__(self, method: Callable):
            self._method = method

        def __get__(self, obj, objtype):
            """Support instance methods."""
            return partial(self.__call__, obj)

        def __call__(self, *args, **kwargs):
            """
            Args:
                method (Callable): Class method.
            Raises:
                AttributeError: live_plot and data_point are required.
                ValueError: Line plots accept exactly x and f values.
                ValueError: Scatter3D plots accept exactly x, y and f values.
            """
            if (
                "live_plot" not in kwargs
                or kwargs["live_plot"] is None
                or "data_packet" not in kwargs
                or kwargs["data_packet"] is None
            ):
                raise AttributeError("live_plot and point info are required.")
            live_plot: LivePlot = kwargs.get("live_plot")
            data_packet: LivePlotPacket = kwargs.get("data_packet")
            if live_plot.plot_type in (LivePlotType.LINES, LivePlotType.SCATTER) and (
                data_packet.data.z is not None or data_packet.data.x is None or data_packet.data.y is None
            ):
                raise ValueError("LINES and SCATTER plots accept exactly x and y values.")
            if live_plot.plot_type == LivePlotType.SCATTER3D and (
                data_packet.data.x is None or data_packet.data.y is None or data_packet.data.z is None
            ):
                raise ValueError("SCATTER3D plots accept exactly x, y and z values.")

            if live_plot.plot_type == LivePlotType.HEATMAP and (
                data_packet.data.x is None or data_packet.data.y is None or data_packet.data.z is None
            ):
                raise ValueError("HEATMAP plots accept exactly x, y and z values.")

            if live_plot.plot_type == LivePlotType.HEATMAP and (
                live_plot.axis.x_axis is None or live_plot.axis.y_axis is None
            ):
                raise ValueError("HEATMAP plots need to have axis provided on live-plot creation.")

            return self._method(*args, **kwargs)

    async def send_data(
        self,
        plot_id: int,
        x: npt.NDArray[np.float_ | np.int_] | list[float] | list[int] | float | int,
        y: npt.NDArray[np.float_ | np.int_] | list[float] | list[int] | float | int,
        z: npt.NDArray[np.float_ | np.int_] | list[float] | list[int] | float | int | None,
    ):
        """
        Send the data corresponding to one or more points over the ws connection associated to the plot represented by
        the provided plot_id
        Args:
            plot_id: id of the plot to send points to
            x: x value of the point(s) to be drawn
            y: y value of the point(s) to be drawn
            z: z value of the point(s) to be drawn
        """
        live_plot = self._get_live_plot(plot_id=plot_id)
        data_packet = LivePlotPacket.build_packet(
            plot_id=live_plot.plot_id,
            plot_type=live_plot.plot_type,
            x=x,
            y=y,
            z=z,
            labels=live_plot.labels,
            axis=live_plot.axis,
        )
        await self._send_data(live_plot=live_plot, data_packet=data_packet)

    @CheckDataAndPlotTypeCompatibility
    def _send_data(self, live_plot: LivePlot, data_packet: LivePlotPacket):
        """Actually sends the data using the live_plot's socket"""
        return live_plot.send_data(data=data_packet)
